from pathlib import PosixPath
from unittest.mock import Mock

import pytest
from rpmlint.cli import process_lint_args
from rpmlint.config import Config
from rpmlint.lint import Lint

from Testing import HAS_CHECKBASHISMS, HAS_DASH, HAS_RPMDB


@pytest.mark.parametrize('test_arguments', [['-c', 'rpmlint/configs/thisdoesntexist.toml']])
def test_parsing_non_existing_config_file(test_arguments):
    with pytest.raises(SystemExit) as exc:
        process_lint_args(test_arguments)
    assert exc.value.code == 2


@pytest.mark.parametrize('test_arguments', [['-c', 'rpmlint/configdefaults.toml']])
def test_parsing_config_file(test_arguments):
    parsed = process_lint_args(test_arguments)

    assert len(parsed['config']) == 1
    assert parsed['config'][0] == PosixPath('rpmlint/configdefaults.toml')


@pytest.mark.skipif(not HAS_CHECKBASHISMS, reason='Optional dependency checkbashisms not installed')
@pytest.mark.skipif(not HAS_DASH, reason='Optional dependency dash not installed')
@pytest.mark.parametrize('test_arguments', [['-c', 'configs/openSUSE']])
def test_parsing_opensuse_conf(test_arguments):
    parsed = process_lint_args(test_arguments)

    assert len(parsed['config']) == 7
    assert PosixPath('configs/openSUSE/opensuse.toml') in parsed['config']
    assert PosixPath('configs/openSUSE/licenses.toml') in parsed['config']
    assert PosixPath('configs/openSUSE/pie-executables.toml') in parsed['config']

    defaultcfg = Config()
    lint = Lint(parsed)
    default_checks = defaultcfg.configuration['Checks']
    checks = lint.config.configuration['Checks']
    # Verify that all original Checks are enabled and some new are added
    for check in default_checks:
        assert check in checks
    assert len(checks) > len(default_checks)

    # Verify that all scoring keys are a known checks
    checks = set(lint.output.error_details.keys())
    checks |= set(defaultcfg.configuration['Descriptions'].keys())

    score_keys = lint.config.configuration['Scoring'].keys()
    for score_key in score_keys:
        if score_key.startswith('percent-in-'):
            continue
        assert score_key in checks


@pytest.mark.skipif(not HAS_CHECKBASHISMS, reason='Optional dependency checkbashisms not installed')
@pytest.mark.skipif(not HAS_DASH, reason='Optional dependency dash not installed')
@pytest.mark.parametrize('test_arguments', [['-c', 'configs/Fedora']])
def test_parsing_fedora_conf(test_arguments):
    parsed = process_lint_args(test_arguments)

    assert len(parsed['config']) == 5
    assert PosixPath('configs/Fedora/fedora.toml') in parsed['config']
    assert PosixPath('configs/Fedora/licenses.toml') in parsed['config']
    assert PosixPath('configs/Fedora/users-groups.toml') in parsed['config']

    defaultcfg = Config()
    lint = Lint(parsed)
    default_checks = defaultcfg.configuration['Checks']
    checks = lint.config.configuration['Checks']
    # Verify that all original Checks are enabled and some new are added
    for check in default_checks:
        assert check in checks
    assert len(checks) > len(default_checks)

    # Verify that all scoring keys are a known checks
    checks = set(lint.output.error_details.keys())
    checks |= set(defaultcfg.configuration['Descriptions'].keys())

    score_keys = lint.config.configuration['Scoring'].keys()
    for score_key in score_keys:
        if score_key.startswith('percent-in-'):
            continue
        assert score_key in checks


def test_reset_check():
    files = ['test/spec/SpecCheck2.spec', 'test/spec/SpecCheck3.spec']
    options = process_lint_args(['--checks', 'SpecCheck'] + files)
    lint = Lint(options)
    lint.run()
    out = lint.output.print_results(lint.output.results, lint.config)
    assert 'more-than-one-%changelog-section' not in out


@pytest.mark.skipif(not HAS_RPMDB, reason='No RPM database present')
@pytest.mark.parametrize('args', [
    ['test/spec/SpecCheck2.spec', 'test/spec/SpecCheck3.spec'],
    ['-i', 'rpm', 'glibc'],
    ['test/spec/SpecCheck2.spec', '-i', 'rpm'],
    ['test/spec/SpecCheck2.spec', 'test/spec/SpecCheck3.spec', '-i', 'rpm', 'glibc'],
])
def test_validate_filters(args):
    options = process_lint_args(args)
    lint = Lint(options)
    lint.output.validate_filters = Mock(wraps=lint.output.validate_filters)
    lint.run()
    lint.output.validate_filters.assert_called_once()



from Testing import get_tested_mock_package


AlternativeConfFolder = get_tested_mock_package(
    lazyload=True,
    header={'requires': [], 'POSTIN': '', 'POSTUN': ''},
    name='alternatives',
    files={
        '/usr/share/libalternatives/rst2html/311.conf': {
            'create_dirs': True,
            'content': 'bin=/usr/bin/rst2html-3.11',
        },
        '/usr/share/libalternatives/rst2html/1313.conf': {
            'create_dirs': True,
            'content': 'binary=/usr/bin/=rst2html-3.13',
        },
        '/usr/share/libalternatives/ldaptor-ldap2dhcpconf/311.conf': {
            'create_dirs': True,
            'content': 'binary=/usr/bin/ldaptor-ldap2dhcpconf-3.11',
        },
        '/usr/share/libalternatives/ldaptor-ldap2dhcpconf/1311.conf': {
            'create_dirs': True,
            'content': 'binary=/usr/bin/ldaptor-ldap2dhcpconf-3.13',
        },
    }
)



from Testing import get_tested_mock_package


SyntaxPackageandPackage = get_tested_mock_package(
    lazyload=True,
    header={
        'requires': [],
        'supplements': ['packageand(c:d)'],
        'recommends': ['packageand(a:b)'],
        'suggests': ['(a and b)'],
        'conflicts': [
            'bee',
            'otherproviders(yast2_theme)',
        ],
    },
)


SyntaxAndPackage = get_tested_mock_package(
    lazyload=True,
    header={
        'requires': [],
        'supplements': ['(c and d)'],
        'recommends': ['b'],
        'suggests': ['(a and b)'],
        'conflicts': [
            'bee',
        ],
    },
)



from unittest.mock import patch

from mockdata.mock_appdata import AppDataPackage
import pytest
from rpmlint.checks.AppDataCheck import AppDataCheck
from rpmlint.filter import Filter

from Testing import CONFIG, HAS_APPSTREAM_GLIB


@pytest.fixture(scope='function', autouse=True)
def appdatacheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = AppDataCheck(CONFIG, output)
    return output, test


@pytest.mark.skipif(not HAS_APPSTREAM_GLIB, reason='Optional dependency appstream-glib not installed')
@pytest.mark.parametrize('package', [AppDataPackage])
def test_appdata_fail(package, appdatacheck):
    output, test = appdatacheck
    test.check(package)
    out = output.print_results(output.results)
    # there are two borked packages
    assert len(output.results) == 2
    assert 'invalid-appdata-file' in out


@pytest.mark.parametrize('package', [AppDataPackage])
@patch('rpmlint.checks.AppDataCheck.AppDataCheck.cmd', 'command-really-not-found')
def test_appdata_fail_no_checker(package, appdatacheck):
    output, test = appdatacheck
    test.check(package)
    out = output.print_results(output.results)
    # there is just one borked file as the other is invalid content
    # but valid xml
    assert len(output.results) == 1
    assert 'invalid-appdata-file' in out



import re

from mockdata.mock_bashisms import BashismsPackage
from mockdata.mock_build_date import BuildDatePackage
import pytest
from rpmlint.checks.BuildRootAndDateCheck import BuildRootAndDateCheck
from rpmlint.filter import Filter

from Testing import CONFIG


@pytest.fixture(scope='function', autouse=True)
def builddatecheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = BuildRootAndDateCheck(CONFIG, output)
    return output, test


@pytest.mark.parametrize('package', [BuildDatePackage])
def test_build_date_time(package, builddatecheck):
    output, test = builddatecheck
    test.istoday = re.compile('Jan  1 2019')
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: file-contains-date-and-time /bin/with-datetime' in out
    assert 'E: file-contains-current-date /bin/with-date' in out


@pytest.mark.parametrize('package', [BashismsPackage])
def test_build_date_time_correct(package, builddatecheck):
    output, test = builddatecheck
    test.istoday = re.compile('Jan  1 2019')
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: file-contains-date-and-time' not in out
    assert 'E: file-contains-current-date' not in out



from mockdata.mock_tmp_files import (
    SystemdTempfilesOkPackage,
    SystemdTempfilesPackage,
    TempfiledPackage,
)
import pytest
from rpmlint.checks.TmpFilesCheck import TmpFilesCheck
from rpmlint.filter import Filter

from Testing import CONFIG


@pytest.fixture(scope='function', autouse=True)
def tmpfilescheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = TmpFilesCheck(CONFIG, output)
    return output, test


@pytest.mark.parametrize('package', [TempfiledPackage])
def test_tmpfiles(package, tmpfilescheck):
    output, test = tmpfilescheck
    test.check(package)
    out = output.print_results(output.results)

    assert 'W: pre-with-tmpfile-creation ' not in out
    assert 'W: post-without-tmpfile-creation /usr/lib/tmpfiles.d/krb5.conf' in out
    assert 'W: tmpfile-not-in-filelist /var/lib/kerberos' in out
    assert 'W: tmpfile-not-regular-file /usr/lib/tmpfiles.d/symlink.conf' in out


@pytest.mark.parametrize('package', [SystemdTempfilesPackage])
def test_tmpfiles2(package, tmpfilescheck):
    output, test = tmpfilescheck
    test.check(package)
    out = output.print_results(output.results)

    assert 'W: pre-with-tmpfile-creation /usr/lib/tmpfiles.d/systemd-tmpfiles.conf' in out
    assert 'W: post-without-tmpfile-creation' in out
    assert 'W: tmpfile-not-in-filelist /run/my_new_directory' in out
    assert 'W: tmpfile-not-regular-file' not in out


@pytest.mark.parametrize('package', [SystemdTempfilesOkPackage])
def test_tmpfiles_correct(package, tmpfilescheck):
    output, test = tmpfilescheck
    test.check(package)
    out = output.print_results(output.results)

    assert 'W: pre-with-tmpfile-creation' not in out
    assert 'W: post-without-tmpfile-creation' not in out
    assert 'W: tmpfile-not-regular-file' not in out
    assert 'W: tmpfile-not-in-filelist' not in out



from Testing import get_tested_mock_package


HardlinksAndDuplicatesPresent = get_tested_mock_package(
    lazyload=True,
    files={
        '/etc/bar': {'content': 'large enough file', 'metadata': {'mode': 33188}},
        '/etc/bar2': {'content': 'large enough file', 'metadata': {'mode': 33188}},
        '/etc/bar3': {'content': 'large enough file', 'metadata': {'mode': 33188}},
        '/etc/foo': {'content': 'Foo file', 'metadata': {'mode': 33188, 'inode': 10}},
        '/etc/foo2': {'content': 'Foo 2 file', 'metadata': {'mode': 33188, 'flags': 1, 'inode': 5}},
        '/etc/small': {'content': ' \n', 'metadata': {'mode': 33188}},
        '/etc/small2': {'content': ' \n', 'metadata': {'mode': 33188}},
        '/etc/strace1.txt': {'content': 'this is a very large file', 'metadata': {'mode': 33188, 'size': 270509}},
        '/etc/strace2.txt': {'content': 'this is a very large file', 'metadata': {'mode': 33188, 'size': 270509}},
        '/var/foo': {'content': 'Foo file', 'metadata': {'mode': 33188, 'inode': 10}},
        '/var/foo2': {'content': 'Foo 2 file', 'metadata': {'mode': 33188, 'flags': 1, 'inode': 5}}
    }
)

NoHardLinksAndDuplicatesPresent = HardlinksAndDuplicatesPresent.clone(
    files={
        '/usr/share/bad-crc.zip': {'content': 'this is a zip file', 'metadata': {'mode': 33188, 'flags': 1}},
        '/usr/share/uncompressed.zip': {'content': 'this is an another zip file', 'metadata': {'mode': 33188, 'flags': 1}},
    },
)



from mockdata.mock_xinetd import RequireXinetd
import pytest
from rpmlint.checks.XinetdDepCheck import XinetdDepCheck
from rpmlint.filter import Filter

from Testing import CONFIG


@pytest.fixture(scope='function', autouse=True)
def xinetdcheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = XinetdDepCheck(CONFIG, output)
    return output, test


@pytest.mark.parametrize('package', [RequireXinetd])
def test_xinetd(package, xinetdcheck):
    output, test = xinetdcheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: obsolete-xinetd-requirement' in out



import concurrent.futures
import contextlib
from pathlib import Path
import re
import stat

from rpmlint.arparser import ArParser
from rpmlint.checks.AbstractCheck import AbstractCheck
from rpmlint.lddparser import LddParser
from rpmlint.objdumpparser import ObjdumpParser
from rpmlint.pkg import FakePkg, InstalledPkg
from rpmlint.readelfparser import ReadelfParser
from rpmlint.stringsparser import StringsParser

KERNEL_MODULES_PATHS = ('/lib/modules/', '/usr/lib/modules/')
GLIBC_EMPTY_ARCHIVES = ('libanl', 'libdl', 'libpthread', 'librt', 'libutil')


class BinariesCheck(AbstractCheck):
    """
    Checks for binary files in the package.
    """
    validso_regex = re.compile(r'(\.so\.\d+(\.\d+)*|\d\.so)$')
    soversion_regex = re.compile(r'.*?(-(?P<pkgversion>[0-9][.0-9]*))?\.so(\.(?P<soversion>[0-9][.0-9]*))?')
    usr_lib_regex = re.compile(r'^/usr/lib(64)?/')
    ldso_soname_regex = re.compile(r'^ld(-linux(-(ia|x86_)64))?\.so')

    numeric_dir_regex = re.compile(r'/usr(?:/share)/man/man./(.*)\.[0-9](?:\.gz|\.bz2)')
    versioned_dir_regex = re.compile(r'[^.][0-9]')
    so_regex = re.compile(r'/lib(64)?/[^/]+\.so(\.[0-9]+)*$')
    bin_regex = re.compile(r'^(/usr(/X11R6)?)?/s?bin/')
    la_file_regex = re.compile(r'\.la$')
    invalid_dir_ref_regex = re.compile(r'/(home|tmp)(\W|$)')
    usr_arch_share_regex = re.compile(r'/share/.*/(?:x86|i.86|x86_64|ppc|ppc64|s390|s390x|ia64|m68k|arm|aarch64|mips|riscv)')
    python_module_regex = re.compile(r'.*\.(\w*(python|pypy)\w*(-\w+){4}|abi3)\.so')
    # Starts with ELF or with a special qualifiers like setuid or setgid
    # https://github.com/rpm-software-management/rpmlint/issues/1088
    elf_regex = re.compile(r'^(\w+ )?ELF ')

    lto_text_like_sections = {'.preinit_array', '.init_array', '.fini_array'}
    # The following sections are part of the RX ABI and do correspond to .text, .data and .bss
    lto_text_like_sections |= {'P', 'D_1', 'B_1'}

    # The list is taken from glibc: sysdeps/${arch}/stackinfo.h
    default_executable_stack_archs = re.compile(r'alpha|arm.*|hppa|i.86|m68k|microblaze|mips|ppc|s390|s390x|sh|sparc|x86_64')

    rpath_origin = '$ORIGIN'
    hpc_locations = ('/usr/lib/mpi/', '/usr/lib64/mpi/', '/usr/lib/hpc/')

    def __init__(self, config, output):
        super().__init__(config, output)
        self.checked_files = 0
        self.system_lib_paths = tuple(config.configuration['SystemLibPaths'])
        self.pie_exec_regex_list = []
        for regex in config.configuration['PieExecutables']:
            self.pie_exec_regex_list.append(re.compile(regex))
        self.usr_lib_exception_regex = re.compile(config.configuration['UsrLibBinaryException'])

        self.setgid_call_regex = self.create_regexp_call(r'set(?:res|e)?gid')
        self.setuid_call_regex = self.create_regexp_call(r'set(?:res|e)?uid')
        self.setgroups_call_regex = self.create_regexp_call(r'(?:ini|se)tgroups')
        self.mktemp_call_regex = self.create_regexp_call('mktemp')
        self.gethostbyname_call_regex = self.create_regexp_call(r'(gethostbyname|gethostbyname2|gethostbyaddr|gethostbyname_r|gethostbyname2_r|gethostbyaddr_r)')

        # register all check functions
        self.check_functions = [self._check_lto_section,
                                self._check_no_text_in_archive,
                                self._check_missing_symtab_in_archive,
                                self._check_missing_debug_info_in_archive,
                                self._check_executable_stack,
                                self._check_shared_library,
                                self._check_dependency,
                                self._check_library_dependency_location,
                                self._check_security_functions,
                                self._check_rpath,
                                self._check_library_dependency,
                                self._check_forbidden_functions,
                                self._check_executable_shlib,
                                self._check_optflags,
                                self._check_hash_sections,
                                self._check_no_patchable_function_entries_in_archive]

    def reset(self):
        self.checked_files = 0

    @staticmethod
    def create_nonlibc_regexp_call(call):
        r = r'(%s)\s?.*$' % call
        return re.compile(r)

    @staticmethod
    def create_regexp_call(call):
        r = r'(%s(?:@GLIBC\S+)?)(?:\s|$)' % call
        return re.compile(r)

    def _check_libtool_wrapper(self, pkg, fname, pkgfile):
        """
        Print an error if the fname file contains a libtool wrapper shell
        script.
        """
        if 'shell script' in pkgfile.magic:
            file_start = None
            with contextlib.suppress(IOError), open(pkgfile.path, 'rb') as inputf:
                file_start = inputf.read(2048)
            if (file_start and b'This wrapper script should never '
                               b'be moved out of the build directory'
                    in file_start):
                self.output.add_info('E', pkg, 'libtool-wrapper-in-package',
                                     fname)

    def _check_invalid_la_file(self, pkg, fname):
        """
        Check if the fname is an .la file and contains a reference to the
        invalid directories ('/tmp' or '/home').

        If so then print a corresponding error with the matching line numbers.
        """
        if self.la_file_regex.search(fname):
            line = pkg.grep(self.invalid_dir_ref_regex, fname)
            if line:
                self.output.add_info('E', pkg, 'invalid-la-file', fname,
                                     f'(line {line})')

    def _check_binary_in_noarch(self, pkg, bin_name):
        """
        Print an error if the binary file bin_name is in the noarch package.
        """
        if pkg.arch == 'noarch':
            self.output.add_info('E', pkg,
                                 'arch-independent-package-contains-binary-or-object',
                                 bin_name)

    def _check_binary_in_usr_share(self, pkg, bin_name):
        """
        Print an error if binary file bin_name is installed in /usr/share.

        We suppose that the package is arch dependent.
        """
        if bin_name.startswith('/usr/share/') and \
                not self.usr_arch_share_regex.search(bin_name):
            self.output.add_info('E', pkg, 'arch-dependent-file-in-usr-share',
                                 bin_name)

    def _check_binary_in_etc(self, pkg, bin_name):
        """
        Print an error if binary file bin_name is installed in /etc directory.

        We suppose that the package is arch dependent.
        """
        if bin_name.startswith('/etc/'):
            self.output.add_info('E', pkg, 'binary-in-etc', bin_name)

    def _check_unstripped_binary(self, bin_name, pkg, pkgfile):
        """
        Print a warning if the bin_name binary has unstripped debug symbols.

        We suppose that the package is arch dependent and bin_name is not
        ocaml native, lua bytecode, .o or .static.
        """
        if 'not stripped' in pkgfile.magic:
            self.output.add_info('W', pkg, 'unstripped-binary-or-object',
                                 bin_name)

    def _check_non_pie(self, pkg, bin_name):
        """
        Check if the bin_name binary is built with PIE.

        Print an error message if it's not while PIE is forced in
        configuration. Print a warning if it's not forced.
        We suppose that the package is arch dependent and bin_name is binary
        executable.
        """
        if not self.is_shobj and not self.is_pie_exec:
            if any(regex.fullmatch(bin_name) for regex in self.pie_exec_regex_list):
                self.output.add_info('E', pkg,
                                     'non-position-independent-executable',
                                     bin_name)
            else:
                self.output.add_info('W', pkg,
                                     'position-independent-executable-suggested',
                                     bin_name)

    def _check_exec_in_library(self, pkg, has_lib, exec_files):
        """
        Check if the library package has an executable file installed.

        Print an error for every such file.
        """
        if has_lib:
            for f in exec_files:
                self.output.add_info('E', pkg, 'executable-in-library-package', f)

    def _check_non_versioned(self, pkg, has_lib, exec_files):
        """
        Check if the library package contains library files in non-versioned
        directories.

        Print an error for every such file.
        """
        if has_lib:
            for f in pkg.files:
                res = self.numeric_dir_regex.search(f)
                fn = res and res.group(1) or f
                if f not in exec_files and not self.so_regex.search(f) and \
                        not self.versioned_dir_regex.search(fn):
                    self.output.add_info('E', pkg, 'non-versioned-file-in-library-package', f)

    def _check_no_binary(self, pkg, has_binary, has_file_in_lib64):
        """
        Check if the arch dependent package contains any binaries.

        Print an error if there is no binary and it's not noarch.
        """
        if not has_binary and not has_file_in_lib64 and \
                pkg.arch != 'noarch':
            self.output.add_info('E', pkg, 'no-binary')

    def _check_noarch_with_lib64(self, pkg, has_file_in_lib64):
        """
        Print an error if we have a noarch package that contains a file
        in /usr/lib64.
        """
        if pkg.arch == 'noarch' and has_file_in_lib64:
            self.output.add_info('E', pkg, 'noarch-with-lib64')

    def _check_only_non_binary_in_usrlib(self, pkg, has_usr_lib_file, has_binary_in_usr_lib):
        """
        Check and print a warning if we have _only_ non-binary files in the
        '/usr/lib'.

        Note: non-binaries allowed via UsrLibBinaryException config option
        are considered binaries.
        """
        if has_usr_lib_file and not has_binary_in_usr_lib:
            self.output.add_info('W', pkg, 'only-non-binary-in-usr-lib')

    def _check_no_text_in_archive(self, pkg, pkgfile):
        """
        For an archive, test if any .text sections is non-empty.
        """
        if self.is_archive:
            for comment in self.readelf_parser.comment_section_info.comments:
                if comment.startswith('GHC '):
                    return

            # Starting with glibc 2.34, some static libraries were moved to libc
            # and there are empty archives for backward compatibility. Skip these
            # libraries.
            stem = Path(pkgfile.name).stem
            if stem in GLIBC_EMPTY_ARCHIVES or (stem.endswith('_p') and stem[:-2] in GLIBC_EMPTY_ARCHIVES):
                return

            for elf_file in self.readelf_parser.section_info.elf_files:
                for section in elf_file:
                    sn = section.name
                    if ((sn in self.lto_text_like_sections or
                         sn == '.fini_array' or sn.startswith('.text') or
                         sn.startswith('.data')) and
                            section.size > 0):
                        return
            self.output.add_info('E', pkg, 'lto-no-text-in-archive', pkgfile.name)
            return

    def _check_no_patchable_function_entries_in_archive(self, pkg, pkgfile):
        """
        For static libraries, we should not ship __patchable_function_entries as it can
        accidentally make a shared library or an executable live-patchable.
        """
        if self.is_archive:
            for elf_file in self.readelf_parser.section_info.elf_files:
                for section in elf_file:
                    if section.name == '__patchable_function_entries':
                        self.output.add_info('E', pkg, 'patchable-function-entry-in-archive', pkgfile.name)
                        return

    def _check_missing_symtab_in_archive(self, pkg, pkgfile):
        """
        FIXME Add test coverage.
        """
        if self.is_archive:
            for elf_file in self.readelf_parser.section_info.elf_files:
                for section in elf_file:
                    if section.name == '.symtab':
                        return

            self.output.add_info('E', pkg, 'static-library-without-symtab', pkgfile.name)

    def _check_missing_debug_info_in_archive(self, pkg, pkgfile):
        if self.is_archive:
            for elf_file in self.readelf_parser.section_info.elf_files:
                for section in elf_file:
                    if section.name.startswith('.debug_'):
                        return
            self.output.add_info('E', pkg, 'static-library-without-debuginfo', pkgfile.name)

    # Check for LTO sections
    def _check_lto_section(self, pkg, pkgfile):
        for elf_file in self.readelf_parser.section_info.elf_files:
            for section in elf_file:
                if '.gnu.lto_.' in section.name:
                    self.output.add_info('E', pkg, 'lto-bytecode', pkgfile.name)
                    return

    def _check_executable_stack(self, pkg, pkgfile):
        """
        Check if the stack is declared as executable which is usually an error.
        """

        # Skip architectures that have non-executable stack by default
        if pkg.arch and not self.default_executable_stack_archs.fullmatch(pkg.arch):
            return

        # Do not check kernel modules and archives
        if not self.is_archive and not any(pkgfile.name.startswith(p) for p in KERNEL_MODULES_PATHS):
            stack_headers = [h for h in self.readelf_parser.program_header_info.headers if h.name == 'GNU_STACK']
            if not stack_headers:
                self.output.add_info('E', pkg, 'missing-PT_GNU_STACK-section', pkgfile.name)
            elif 'E' in stack_headers[0].flags:
                self.output.add_info('E', pkg, 'executable-stack', pkgfile.name)

    def _check_soname_symlink(self, pkg, shlib, soname):
        """
        Check that we have a symlink with the soname in the package and it
        points to the checked shared library.

        Print an error if the symlink is invalid or missing.
        """
        path = Path(shlib)
        symlink = path.parent / soname
        try:

            link = pkg.files[str(symlink)].linkto
            if link not in (shlib, path.name, ''):
                self.output.add_info('E', pkg, 'invalid-ldconfig-symlink', shlib, link)
        except KeyError:
            # if we do not have a symlink, report an issue
            if path.name.startswith('lib') or path.name.startswith('ld-'):
                self.output.add_info('E', pkg, 'no-ldconfig-symlink', shlib)

    def _check_shared_library(self, pkg, pkgfile):
        """
        Various checks for the shared library.

        1) Print 'no-soname' warning it the library has no soname present.
        2) Print 'invalid-soname' error if the soname is not valid.
        3) Print 'shlib-policy-name-error' error when the library major
           version is not present in the package name.
        4) Print 'shlib-with-non-pic-code' error if the library contains
           object code that was compiled without -fPIC.
        """
        if not self.readelf_parser.is_shlib:
            return

        soname = self.readelf_parser.dynamic_section_info.soname
        if not soname:
            self.output.add_info('W', pkg, 'no-soname', pkgfile.name)
        else:
            if not self.validso_regex.search(soname):
                self.output.add_info('E', pkg, 'invalid-soname', pkgfile.name, soname)
            else:
                self._check_soname_symlink(pkg, pkgfile.name, soname)

                # check if the major version of the library is in the package
                # name (check only for lib* packages)
                # for now skip all HPC libraries (#901)
                if pkg.name.startswith('lib') and not pkgfile.name.startswith(self.hpc_locations):
                    # SLPP is defined here: https://en.opensuse.org/openSUSE:Shared_library_packaging_policy#Package_naming
                    # Example:
                    # SONAME = libgame2-1.9.so.10.0.0
                    # expected package name: libgame2-1_9-10_0_0
                    res = self.soversion_regex.search(soname)
                    if res:
                        parts = [x.replace('.', '_') for x in (res.group('pkgversion'), res.group('soversion')) if x]
                        soversion = '-'.join(parts)
                        pkgname = pkg.name
                        if '.' in pkgname:
                            pkgname = pkgname[:pkgname.rindex('.')]
                        if soversion and not pkgname.endswith(soversion):
                            self.output.add_info('E', pkg, 'shlib-policy-name-error',
                                                 f'SONAME: {soname} ({pkgfile.name}), expected package suffix: {soversion}')

        # check if the object code in the library is compiled with PIC
        if self.readelf_parser.dynamic_section_info['TEXTREL']:
            self.output.add_info('E', pkg, 'shlib-with-non-pic-code', pkgfile.name)

    def _check_dependency(self, pkg, pkgfile):
        """
        FIXME Add test coverage.
        """
        # Undefined symbol and unused direct dependency checks make sense only
        # for installed packages.
        # skip debuginfo: https://bugzilla.redhat.com/190599
        #
        # following issues are errors for shared libs and warnings for executables
        if not self.is_dynamically_linked:
            return

        # Skip python packages
        if self.python_module_regex.fullmatch(pkgfile.name):
            return

        if not self.is_archive and not self.readelf_parser.is_debug:
            info_type = 'E' if self.readelf_parser.is_shlib else 'W'
            for symbol in self.ldd_parser.undefined_symbols:
                self.output.add_info(info_type, pkg, 'undefined-non-weak-symbol', pkgfile.name, symbol)
            for dependency in self.ldd_parser.unused_dependencies:
                self.output.add_info(info_type, pkg, 'unused-direct-shlib-dependency',
                                     pkgfile.name, dependency)

    def _check_library_dependency_location(self, pkg, pkgfile):
        """
        FIXME Add test coverage.
        """
        if not self.is_dynamically_linked:
            return

        if not self.is_archive:
            for dependency in self.ldd_parser.dependencies:
                if dependency.startswith('/opt/'):
                    self.output.add_info('E', pkg, 'linked-against-opt-library', pkgfile.name, dependency)
                    break

        nonusr = ('/bin', '/lib', '/sbin')
        if pkgfile.name.startswith(nonusr):
            for dependency in self.ldd_parser.dependencies:
                if dependency.startswith('/usr/'):
                    self.output.add_info('W', pkg, 'linked-against-usr-library', pkgfile.name, dependency)
                    break

    def _check_security_functions(self, pkg, pkgfile):
        setgid = any(self.readelf_parser.symbol_table_info.get_functions_for_regex(self.setgid_call_regex))
        setuid = any(self.readelf_parser.symbol_table_info.get_functions_for_regex(self.setuid_call_regex))
        setgroups = any(self.readelf_parser.symbol_table_info.get_functions_for_regex(self.setgroups_call_regex))
        mktemp = any(self.readelf_parser.symbol_table_info.get_functions_for_regex(self.mktemp_call_regex))
        gethostbyname = any(self.readelf_parser.symbol_table_info.get_functions_for_regex(self.gethostbyname_call_regex))

        if setgid and setuid and not setgroups:
            is_uid = stat.S_ISUID & pkgfile.mode
            self.output.add_info('W' if is_uid else 'E', pkg, 'missing-call-to-setgroups-before-setuid', pkgfile.name)

        if mktemp:
            self.output.add_info('E', pkg, 'call-to-mktemp', pkgfile.name)

        if gethostbyname:
            self.output.add_info('W', pkg, 'binary-or-shlib-calls-gethostbyname', pkgfile.name)

    def _check_rpath(self, pkg, pkgfile):
        for runpaths in self.readelf_parser.dynamic_section_info.runpaths:
            for runpath in runpaths.split(':'):
                if self.rpath_origin in runpath:
                    runpath = runpath.replace(self.rpath_origin, str(Path(pkgfile.name).parent))
                    runpath = str(Path(runpath).resolve())
                if not runpath.startswith(self.system_lib_paths) and not self.usr_lib_regex.search(runpath):
                    self.output.add_info('E', pkg, 'binary-or-shlib-defines-rpath', pkgfile.name, f'(RUNPATH: {runpaths})')
                    return

    def _check_library_dependency(self, pkg, pkgfile):
        if (self.is_archive or
                any(pkgfile.name.startswith(p) for p in KERNEL_MODULES_PATHS) or
                self.python_module_regex.fullmatch(pkgfile.name)):
            return

        dyn_section = self.readelf_parser.dynamic_section_info
        if not len(dyn_section.needed) and not (dyn_section.soname and
                                                self.ldso_soname_regex.search(dyn_section.soname)):
            if not self.is_shobj:
                self.output.add_info('E', pkg, 'statically-linked-binary', pkgfile.name)
        else:
            # linked against libc ?
            if 'libc.' not in dyn_section.runpaths and \
               (not dyn_section.soname or
                ('libc.' not in dyn_section.soname and
                 not self.ldso_soname_regex.search(dyn_section.soname))):
                for lib in dyn_section.needed:
                    if 'libc.' in lib:
                        return
                if not self.is_shobj:
                    self.output.add_info('W', pkg, 'program-not-linked-against-libc', pkgfile.name)

    def _check_forbidden_functions(self, pkg, pkgfile):
        forbidden_functions = self.config.configuration['WarnOnFunction']
        if forbidden_functions:
            for name, func in forbidden_functions.items():
                # precompile regexps
                f_name = func['f_name']
                func['f_regex'] = self.create_nonlibc_regexp_call(f_name)
                if 'good_param' in func and func['good_param']:
                    func['waiver_regex'] = re.compile(func['good_param'])
                # register descriptions
                self.output.error_details.update({name: func['description']})

        forbidden_calls = []
        for r_name, func in forbidden_functions.items():
            if any(self.readelf_parser.symbol_table_info.get_functions_for_regex(func['f_regex'])):
                forbidden_calls.append(r_name)

        if not forbidden_calls:
            return

        strings_parser = StringsParser(pkgfile.path)
        failed_reason = strings_parser.parsing_failed_reason
        if failed_reason:
            self.output.add_info('E', pkg, 'strings-failed', pkgfile.name, failed_reason)
            return

        forbidden_functions_filtered = []

        for fn in forbidden_calls:
            f = forbidden_functions[fn]
            if 'waiver_regex' not in f:
                forbidden_functions_filtered.append(fn)
                continue

            # See: https://github.com/PyCQA/flake8-bugbear/issues/269
            waiver = any(f['waiver_regex'].search(string) for string in strings_parser.strings)
            if not waiver:
                forbidden_functions_filtered.append(fn)

        for fn in forbidden_functions_filtered:
            self.output.add_info('W', pkg, fn, pkgfile.name, forbidden_functions[fn]['f_name'])

    def _check_executable_shlib(self, pkg, pkgfile):
        if not (pkgfile.mode & stat.S_IEXEC) and self.readelf_parser.is_shlib:
            self.output.add_info('E', pkg, 'shared-library-not-executable', pkgfile.name)

    def _check_optflags(self, pkg, pkgfile):
        if self.is_archive:
            return

        mandatory_optflags = self.config.configuration['MandatoryOptflags']
        forbidden_optflags = self.config.configuration['ForbiddenOptflags']
        if not mandatory_optflags and not forbidden_optflags:
            return

        for dwarf_unit in self.objdump_parser.compile_units:
            tokens = dwarf_unit['producer'].split(' ')
            missing = [mo for mo in mandatory_optflags if mo not in tokens]
            forbidden = [f for f in forbidden_optflags if f in tokens]
            if missing:
                self.output.add_info('W', pkg, 'missing-mandatory-optflags', pkgfile.name, ' '.join(missing))
            if forbidden:
                self.output.add_info('E', pkg, 'forbidden-optflags', pkgfile.name, ' '.join(forbidden))

    def _check_hash_sections(self, pkg, pkgfile):
        if not self.readelf_parser.is_shlib:
            return

        for elf_file in self.readelf_parser.section_info.elf_files:
            needle = {'.hash', '.gnu.hash'}
            for section in elf_file:
                if not needle:
                    break
                if section.name in needle:
                    needle.remove(section.name)

            if '.hash' in needle:
                self.output.add_info('E', pkg, 'missing-hash-section', pkgfile.name)
            if '.gnu.hash' in needle:
                self.output.add_info('W', pkg, 'missing-gnu-hash-section', pkgfile.name)

    def _is_standard_archive(self, pkg, pkgfile):
        # skip Klee bytecode archives
        if pkgfile.path.endswith('.bca'):
            return False

        # return false for e.g. Rust or Go packages that are archives
        # but files in the archive are not an ELF container
        ar_parser = ArParser(pkgfile.path)
        failed_reason = ar_parser.parsing_failed_reason
        if failed_reason:
            self.output.add_info('E', pkg, 'ar-failed', pkgfile.name, failed_reason)
            return False

        needles = ('__.PKGDEF', '_go_.o', 'lib.rmeta')
        return not any(needle for needle in needles if needle in ar_parser.objects)

    def _detect_attributes(self, magic):
        self.is_exec = 'executable' in magic
        self.is_shobj = 'shared object' in magic
        self.is_archive = 'current ar archive' in magic
        self.is_dynamically_linked = 'dynamically linked' in magic
        self.is_pie_exec = 'pie executable' in magic
        self.is_nonstandard_archive = False

    def run_elf_checks(self, pkg, pkgfile):
        if self.is_archive and not self._is_standard_archive(pkg, pkgfile):
            self.is_nonstandard_archive = True
            return

        self.readelf_parser = ReadelfParser(pkgfile.path, pkgfile.name)
        failed_reason = self.readelf_parser.parsing_failed_reason()
        if failed_reason:
            self.output.add_info('E', pkg, 'readelf-failed', pkgfile.name, failed_reason)
            return

        if not self.is_archive:
            if self.is_dynamically_linked:
                is_installed_pkg = isinstance(pkg, (InstalledPkg, FakePkg))
                self.ldd_parser = LddParser(pkgfile.path, pkgfile.name, is_installed_pkg)
                failed_reason = self.ldd_parser.parsing_failed_reason
                if failed_reason:
                    self.output.add_info('E', pkg, 'ldd-failed', pkgfile.name, failed_reason)
                    return

            if (self.config.configuration['MandatoryOptflags'] or
                    self.config.configuration['ForbiddenOptflags']):
                self.objdump_parser = ObjdumpParser(pkgfile.path, pkgfile.name)
                failed_reason = self.objdump_parser.parsing_failed_reason
                if failed_reason:
                    self.output.add_info('E', pkg, 'objdump-failed', pkgfile.name, failed_reason)
                    return

        # NOTE: the speed benefit of the ThreadPoolExecutor is limited due to
        # Global Interpreter Lock (GIL).
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = []
            for fn in self.check_functions:
                futures.append(executor.submit(fn, pkg, pkgfile))
            concurrent.futures.wait(futures)
            for future in futures:
                err = future.exception()
                if err:
                    raise err

    def check_binary(self, pkg):
        exec_files = []
        pkg_has_lib = False
        pkg_has_binary = False
        pkg_has_binary_in_usrlib = False
        pkg_has_usrlib_file = False
        pkg_has_file_in_lib64 = False

        #  go through the all files, run files checks and collect data that are
        #  needed later
        for fname, pkgfile in pkg.files.items():
            # Common tests first
            self._check_libtool_wrapper(pkg, fname, pkgfile)
            self._check_invalid_la_file(pkg, fname)

            # consider non-binary in /usr/lib/ that is allowed by
            # UsrLibBinaryException config option as a "fake" binary and
            # do not throw 'only-non-binary-in-usr-lib' warning then
            if not stat.S_ISDIR(pkgfile.mode) and self.usr_lib_regex.search(fname):
                pkg_has_usrlib_file = True
                if not pkg_has_binary_in_usrlib and \
                        self.usr_lib_exception_regex.search(fname):
                    # Fake that we have binaries there to avoid
                    # only-non-binary-in-usr-lib false positives
                    pkg_has_binary_in_usrlib = True

            # find out if we have a file in /usr/lib64/ directory (needed later
            # for the package checks)
            if fname.startswith(('/usr/lib64', '/lib64')):
                pkg_has_file_in_lib64 = True

            # skip the rest of the tests for non-binaries
            # binary files only from here on
            is_ocaml_native = 'Objective caml native' in pkgfile.magic
            is_lua_bytecode = 'Lua bytecode' in pkgfile.magic
            #  eBPF binaries are arch independent
            # https://github.com/rpm-software-management/rpmlint/issues/1193
            is_ebpf = 'eBPF' in pkgfile.magic
            # Look for ELF in the file magic to check if it's really a binary
            # file
            is_elf = self.elf_regex.match(pkgfile.magic) and not is_ebpf

            if not (is_elf or
                    'current ar archive' in pkgfile.magic or
                    is_ocaml_native or is_lua_bytecode):
                continue

            self.checked_files += 1

            # mark this package as a one that has binary file
            pkg_has_binary = True

            # if there is a binary in /usr/lib then mark this package
            # accordingly
            if pkg_has_usrlib_file and not pkg_has_binary_in_usrlib and \
                    self.usr_lib_regex.search(fname):
                pkg_has_binary_in_usrlib = True

            self._check_binary_in_noarch(pkg, fname)

            # skip the rest of the tests for noarch packages
            # arch dependent packages only from here on
            if pkg.arch == 'noarch':
                continue

            self._check_binary_in_usr_share(pkg, fname)
            self._check_binary_in_etc(pkg, fname)

            # skip the rest of the tests for ocaml native, Lua bytecode,
            # Go .go and .gox, .o and .static
            if is_ocaml_native or is_lua_bytecode or fname.endswith('.o') or \
                    fname.endswith('.static') or fname.endswith('.gox') or \
                    fname.endswith('.go'):
                continue

            self._check_unstripped_binary(fname, pkg, pkgfile)

            # Detect attributes of an ELF file
            self._detect_attributes(pkgfile.magic)

            # run ELF checks
            self.run_elf_checks(pkg, pkgfile)

            if self.is_nonstandard_archive:
                continue

            # inspect binary file
            is_shlib = self.readelf_parser.is_shlib

            if is_shlib:
                pkg_has_lib = True

            # skip non-exec and non-SO
            # executables and shared objects only from here on
            if not self.is_exec and not self.is_shobj:
                continue

            if self.is_shobj and not self.is_exec and '.so' not in fname and \
                    self.bin_regex.search(fname):
                # pkgfile.magic does not contain 'executable' for PIEs
                self.is_exec = True

            if self.is_exec:
                # add to the list of the all exec files
                if self.bin_regex.search(fname):
                    exec_files.append(fname)

                self._check_non_pie(pkg, fname)

        # run checks for the whole package
        # it uses data collected in the previous for-cycle
        self._check_exec_in_library(pkg, pkg_has_lib, exec_files)
        self._check_non_versioned(pkg, pkg_has_lib, exec_files)
        self._check_no_binary(pkg, pkg_has_binary, pkg_has_file_in_lib64)
        self._check_noarch_with_lib64(pkg, pkg_has_file_in_lib64)
        self._check_only_non_binary_in_usrlib(pkg, pkg_has_usrlib_file,
                                              pkg_has_binary_in_usrlib)



import pytest
from rpmlint.checks.SourceCheck import SourceCheck
from rpmlint.filter import Filter

from Testing import CONFIG, get_tested_package


@pytest.fixture(scope='function', autouse=True)
def sourcescheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = SourceCheck(CONFIG, output)
    return output, test


@pytest.mark.parametrize('package', ['source/wrongsrc'])
def test_extension_and_permissions(tmp_path, package, sourcescheck):
    output, test = sourcescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)

    assert len(output.results) == 1

    assert 'inconsistent-file-extension' in out
    assert 'name extension indicates a different compression format' in out

    assert 'strange-permission' not in out
    assert 'a file should have' not in out


@pytest.mark.parametrize('package', ['source/not-compressed-multi-spec'])
def test_compression_and_multispec(tmp_path, package, sourcescheck):
    output, test = sourcescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)

    assert 'source-not-compressed' in out
    assert 'source archive or file in your package is not compressed' in out

    assert 'multiple-specfiles' in out
    assert 'package contains multiple spec files' in out



from pathlib import Path
import stat

from rpm import expandMacro
from rpmlint.checks import FilesCheck
from rpmlint.checks.AbstractCheck import AbstractCheck
from rpmlint.pkg import FakePkg


class LibraryDependencyCheck(AbstractCheck):
    def __init__(self, config, output):
        super().__init__(config, output)
        self.package_requires = {}
        self.package_so_symlinks = {}
        self.package_so_files = {}
        self.package_arch_mapping = {}
        self.isa = expandMacro('%{_isa}')

    def reset(self):
        self.package_requires = {}
        self.package_so_symlinks = {}
        self.package_so_files = {}
        self.package_arch_mapping = {}
        self.isa = expandMacro('%{_isa}')

    def check_binary(self, pkg):
        if pkg.is_source:
            return

        is_devel = FilesCheck.devel_regex.search(pkg.name)
        if is_devel:
            self._process_devel_package(pkg, is_devel)
        else:
            self._process_nondevel_package(pkg)

    def _process_devel_package(self, pkg, is_devel):
        self.package_requires[pkg.name] = [req[0] for req in pkg.requires + pkg.prereq]
        self.package_so_symlinks[pkg.name] = []
        self.package_arch_mapping[pkg.name] = pkg.arch

        for pkgfile in pkg.files.values():
            if stat.S_ISLNK(pkgfile.mode) and pkgfile.name.endswith('.so'):
                link = Path(pkgfile.name).parent / pkgfile.linkto
                self.package_so_symlinks[pkg.name].append(str(link))

    def _process_nondevel_package(self, pkg):
        for pkgfile in pkg.files.values():
            if '.so' in pkgfile.name:
                self.package_so_files[pkgfile.name] = pkg.name

    def after_checks(self):
        for pkgname, so_symlinks in self.package_so_symlinks.items():
            for link in so_symlinks:
                with FakePkg(pkgname) as pkg:
                    pkg.arch = self.package_arch_mapping[pkgname]
                    if link not in self.package_so_files:
                        self.output.add_info('E', pkg, 'no-library-dependency-for', link)
                        break
                    else:
                        definition = self.package_so_files[link]

                        if (definition not in self.package_requires[pkgname] and
                                definition + self.isa not in self.package_requires[pkgname]):
                            self.output.add_info('E', pkg, 'no-library-dependency-on', definition, link)
                            break



import rpm


class PkgFile:
    __slots__ = ['name', 'path', 'flags', 'mode', 'user', 'group', 'linkto',
                 'size', 'md5', 'mtime', 'rdev', 'inode', 'requires', 'provides',
                 'lang', 'magic', 'filecaps']

    def __init__(self, name):
        self.name = name
        # Real path to the file (taking extract dir into account)
        self.path = name
        self.flags = 0
        self.mode = 0
        self.user = None
        self.group = None
        self.linkto = ''
        self.size = None
        self.md5 = None
        self.mtime = 0
        self.rdev = ''
        self.inode = 0
        self.requires = []
        self.provides = []
        self.lang = ''
        self.magic = ''
        self.filecaps = None

    @property
    def is_config(self):
        return self.flags & rpm.RPMFILE_CONFIG

    @property
    def is_doc(self):
        return self.flags & rpm.RPMFILE_DOC

    @property
    def is_noreplace(self):
        return self.flags & rpm.RPMFILE_NOREPLACE

    @property
    def is_ghost(self):
        return self.flags & rpm.RPMFILE_GHOST

    @property
    def is_missingok(self):
        return self.flags & rpm.RPMFILE_MISSINGOK



import re

import rpm
from rpmlint.checks.AbstractCheck import AbstractCheck


class LSBCheck(AbstractCheck):
    """
    Validate that package name, version and release number are LSB compliant.

    The rules are the intersection of compatible NVRs between RPM v3 and DPKG
    for supporting portability across RPM and Debian systems through tools
    like alien.

    Note: It uses values gained from rpm (RPMTAGs) not parsed from .rpm
    filename.
    """
    name_regex = re.compile('^[a-z0-9.+-]+$')
    version_regex = re.compile('^[a-zA-Z0-9.+]+$')

    def check(self, pkg):
        self._check_lsb_name(pkg)
        self._check_lsb_version(pkg)
        self._check_lsb_release(pkg)

    def _check_lsb_name(self, pkg):
        """
        Check if the package name is LSB compliant (only lowercase
        letters, numbers, '.', '+' or '-' characters).
        """
        name = pkg.name
        if name and not self.name_regex.search(name):
            self.output.add_info('E', pkg, 'non-lsb-compliant-package-name', pkg.name)

    def _check_lsb_version(self, pkg):
        """
        Check if the package version number is LSB compliant (only alphanumeric
        symbols, '.' or '+' characters).
        """
        version = pkg[rpm.RPMTAG_VERSION]
        if version and not self.version_regex.search(version):
            self.output.add_info('E', pkg, 'non-lsb-compliant-version', pkg[rpm.RPMTAG_VERSION])

    def _check_lsb_release(self, pkg):
        """
        Check if the package release number is LSB compliant (only alphanumeric
        symbols, '.' or '+' characters).
        """
        release = pkg[rpm.RPMTAG_RELEASE]
        if release and not self.version_regex.search(release):
            self.output.add_info('E', pkg, 'non-lsb-compliant-release', pkg[rpm.RPMTAG_RELEASE])



#############################################################################
# File          : FilesCheck.py
# Package       : rpmlint
# Author        : Frederic Lepied
# Created on    : Mon Oct  4 19:32:49 1999
# Purpose       : test various aspects on files: locations, owner, groups,
#                 permission, setuid, setgid...
#############################################################################

from datetime import datetime
import os
from pathlib import Path
import re
import stat

import rpm
from rpmlint.checks.AbstractCheck import AbstractCheck
from rpmlint.helpers import byte_to_string
from rpmlint.pkg import is_utf8, is_utf8_bytestr

# must be kept in sync with the filesystem package
STANDARD_DIRS = (
    '/',
    '/bin',
    '/boot',
    '/etc',
    '/etc/X11',
    '/etc/opt',
    '/etc/profile.d',
    '/etc/skel',
    '/etc/xinetd.d',
    '/home',
    '/lib',
    '/lib/modules',
    '/lib64',
    '/media',
    '/mnt',
    '/mnt/cdrom',
    '/mnt/disk',
    '/mnt/floppy',
    '/opt',
    '/proc',
    '/root',
    '/run',
    '/sbin',
    '/selinux',
    '/srv',
    '/sys',
    '/tmp',
    '/usr',
    '/usr/X11R6',
    '/usr/X11R6/bin',
    '/usr/X11R6/doc',
    '/usr/X11R6/include',
    '/usr/X11R6/lib',
    '/usr/X11R6/lib64',
    '/usr/X11R6/man',
    '/usr/X11R6/man/man1',
    '/usr/X11R6/man/man2',
    '/usr/X11R6/man/man3',
    '/usr/X11R6/man/man4',
    '/usr/X11R6/man/man5',
    '/usr/X11R6/man/man6',
    '/usr/X11R6/man/man7',
    '/usr/X11R6/man/man8',
    '/usr/X11R6/man/man9',
    '/usr/X11R6/man/mann',
    '/usr/bin',
    '/usr/bin/X11',
    '/usr/etc',
    '/usr/games',
    '/usr/include',
    '/usr/lib',
    '/usr/lib/X11',
    '/usr/lib/games',
    '/usr/lib/gcc-lib',
    '/usr/lib/menu',
    '/usr/lib64',
    '/usr/lib64/gcc-lib',
    '/usr/local',
    '/usr/local/bin',
    '/usr/local/doc',
    '/usr/local/etc',
    '/usr/local/games',
    '/usr/local/info',
    '/usr/local/lib',
    '/usr/local/lib64',
    '/usr/local/man',
    '/usr/local/man/man1',
    '/usr/local/man/man2',
    '/usr/local/man/man3',
    '/usr/local/man/man4',
    '/usr/local/man/man5',
    '/usr/local/man/man6',
    '/usr/local/man/man7',
    '/usr/local/man/man8',
    '/usr/local/man/man9',
    '/usr/local/man/mann',
    '/usr/local/sbin',
    '/usr/local/share',
    '/usr/local/share/man',
    '/usr/local/share/man/man1',
    '/usr/local/share/man/man2',
    '/usr/local/share/man/man3',
    '/usr/local/share/man/man4',
    '/usr/local/share/man/man5',
    '/usr/local/share/man/man6',
    '/usr/local/share/man/man7',
    '/usr/local/share/man/man8',
    '/usr/local/share/man/man9',
    '/usr/local/share/man/mann',
    '/usr/local/src',
    '/usr/sbin',
    '/usr/share',
    '/usr/share/dict',
    '/usr/share/doc',
    '/usr/share/icons',
    '/usr/share/info',
    '/usr/share/man',
    '/usr/share/man/man1',
    '/usr/share/man/man2',
    '/usr/share/man/man3',
    '/usr/share/man/man4',
    '/usr/share/man/man5',
    '/usr/share/man/man6',
    '/usr/share/man/man7',
    '/usr/share/man/man8',
    '/usr/share/man/man9',
    '/usr/share/man/mann',
    '/usr/share/misc',
    '/usr/src',
    '/usr/tmp',
    '/var',
    '/var/cache',
    '/var/db',
    '/var/lib',
    '/var/lib/games',
    '/var/lib/misc',
    '/var/lib/rpm',
    '/var/local',
    '/var/log',
    '/var/mail',
    '/var/nis',
    '/var/opt',
    '/var/preserve',
    '/var/spool',
    '/var/spool/mail',
    '/var/tmp',
)

compressions = r'\.(gz|z|Z|zip|bz2|lzma|xz|zst)'
sub_bin_regex = re.compile(r'^(/usr)?/s?bin/\S+/')
backup_regex = re.compile(r'(~|\#[^/]+\#|((\.orig|\.rej)(' + compressions + ')?))$')
compr_regex = re.compile(compressions + r'$')
absolute_regex = re.compile(r'^/([^/]+)')
absolute2_regex = re.compile(r'^/?([^/]+)')
points_regex = re.compile(r'^\.\./(.*)')
doc_regex = re.compile(r'^/usr(/share|/X11R6)?/(doc|man|info)/|^/usr/share/gnome/help')
bin_regex = re.compile(r'^/(?:usr/(?:s?bin|games)|s?bin)/(.*)')
includefile_regex = re.compile(r'\.(c|h)(pp|xx)?$', re.IGNORECASE)
develfile_regex = re.compile(r'\.(a|cmxa?|mli?|gir)$')
buildconfigfile_regex = re.compile(r'(\.pc|/bin/.+-config)$')
# room for improvement with catching more -R, but also for false positives...
buildconfig_rpath_regex = re.compile(r'(?:-rpath|Wl,-R)\b')
sofile_regex = re.compile(r'/lib(64)?/(.+/)?lib[^/]+\.so$')
devel_regex = re.compile(r'(.*)-(debug(info|source)?|devel|headers|source|static|prof)$')
debuginfo_package_regex = re.compile(r'-debug(info)?$')
debugsource_package_regex = re.compile(r'-debugsource$')
lib_regex = re.compile(r'/lib(?:64)?/lib[A-Za-z0-9](?:(?:|[\w\-\.]*[A-Za-z0-9])\.so\.[\w+\.]+|\w*-\d(?:|[\w\-\.]*[A-Za-z0-9])\.so)$')  # see commit log for detail
ldconfig_regex = re.compile(r'^[^#]*ldconfig', re.MULTILINE)
depmod_regex = re.compile(r'^[^#]*depmod', re.MULTILINE)
install_info_regex = re.compile(r'^[^#]*install-info', re.MULTILINE)
perl_temp_file_regex = re.compile(r'.*perl.*/(\.packlist|perllocal\.pod)$')
scm_regex = re.compile(
    r'/(?:RCS|CVS)/[^/]+$|/\.(?:bzr|cvs|git|hg|svn)ignore$|'
    r',v$|/\.hgtags$|/\.(?:bzr|git|hg|svn)/|/(?:\.arch-ids|{arch})/')
games_path_regex = re.compile(r'^/usr(/lib(64)?)?/games/')
logrotate_regex = re.compile(r'^/etc/logrotate\.d/(.*)')
kernel_modules_regex = re.compile(r'^(?:/usr)/lib/modules/([0-9]+\.[0-9]+\.[0-9]+[^/]*?)/')
kernel_package_regex = re.compile(r'^kernel(22)?(-)?(smp|enterprise|bigmem|secure|BOOT|i686-up-4GB|p3-smp-64GB)?')
normal_zero_length_regex = re.compile(r'^/etc/security/console\.apps/'
                                      r'|/\.nosearch$'
                                      r'|/__init__\.py$'
                                      r'|/py\.typed$'  # https://www.python.org/dev/peps/pep-0561/#packaging-type-information
                                      r'|\.dist-info/REQUESTED$'  # https://www.python.org/dev/peps/pep-0376/#requested
                                      r'|/gem\.build_complete$')
perl_regex = re.compile(r'^/usr/lib/perl5/(?:vendor_perl/)?([0-9]+\.[0-9]+)\.([0-9]+)/')
python_regex = re.compile(r'^/usr/lib(?:64)?/python([.0-9]+)/')
python_bytecode_regex_pep3147 = re.compile(r'^(.*)/__pycache__/(.*?)\.([^.]+)(\.opt-[12])?\.py[oc]$')
python_bytecode_regex = re.compile(r'^(.*)(\.py[oc])$')
log_regex = re.compile(r'^/var/log/[^/]+$')
lib_path_regex = re.compile(r'^(/usr(/X11R6)?)?/lib(64)?')
lib_package_regex = re.compile(r'^(lib|.+-libs)')
hidden_file_regex = re.compile(r'/\.[^/]*$')
manifest_perl_regex = re.compile(r'^/usr/share/doc/perl-.*/MANIFEST(\.SKIP)?$')
shebang_regex = re.compile(br'^#!\s*(\S+)(.*?)$', re.M)
interpreter_regex = re.compile(r'^/(?:usr/)?(?:s?bin|games|libexec(?:/.+)?|(?:lib(?:64)?|share)/.+)/([^/]+)$')
script_regex = re.compile(r'^/((usr/)?s?bin|etc/(rc\.d/init\.d|X11/xinit\.d|cron\.(hourly|daily|monthly|weekly)))/')
sourced_script_regex = re.compile(r'^/etc/(bash_completion\.d|profile\.d)/')
filesys_packages = ['filesystem']  # TODO: make configurable?
quotes_regex = re.compile(r'[\'"]+')
start_certificate_regex = re.compile(r'^-----BEGIN CERTIFICATE-----$')
start_private_key_regex = re.compile(r'^----BEGIN PRIVATE KEY-----$')
non_readable_regexs = (re.compile(r'^/var/log/'),
                       re.compile(r'^/etc/(g?shadow-?|securetty)$'))

man_base_regex = re.compile(r'^/usr(?:/share)?/man(?:/overrides)?/man(?P<category>[^/]+)/(?P<filename>((?P<binary>[^.]+)\..+))')

fsf_license_regex = re.compile(br'(GNU((\s+(Library|Lesser|Affero))?(\s+General)?\s+Public|\s+Free\s+Documentation)\s+Licen[cs]e|(GP|FD)L)', re.IGNORECASE)
fsf_wrong_address_regex = re.compile(br'(675\s+Mass\s+Ave|59\s+Temple\s+Place|02139|51\s+Franklin\s+St)', re.IGNORECASE)

scalable_icon_regex = re.compile(r'^/usr(?:/local)?/share/icons/.*/scalable/')
tcl_regex = re.compile(r'^/usr/lib(64)?/([^/]+/)?pkgIndex\.tcl')

printable_extended_ascii = b'\n\r\t\f\b'
printable_extended_ascii += bytes(range(32, 256))


# See Python sources for a full list of the values here.
# https://github.com/python/cpython/blob/master/Lib/importlib/_bootstrap_external.py
# https://github.com/python/cpython/blob/2.7/Python/import.c
# https://github.com/python/cpython/commit/93602e3af70d3b9f98ae2da654b16b3382b68d50
_python_magic_values = {
    '2.2': [60717],
    '2.3': [62011],
    '2.4': [62061],
    '2.5': [62131],
    '2.6': [62161],
    '2.7': [62211],
    '3.0': [3130],
    '3.1': [3150],
    '3.2': [3180],
    '3.3': [3230],
    '3.4': [3310],
    '3.5': [3350, 3351],  # 3350 for < 3.5.2
    '3.6': [3379],
    '3.7': [3390, 3391, 3392, 3393, 3394],
}


def get_expected_pyc_magic(path, python_default_version):
    """
    .pyc/.pyo files embed a 4-byte magic value identifying which version of
    the python bytecode ABI they are for. Given a path to a .pyc/.pyo file,
    return a (magic ABI values, python version) tuple.  For example,
    '/usr/lib/python3.1/foo.pyc' should return (3151, '3.1').
    The first value will be None if the python version was not resolved
    from the given pathname and the PythonDefaultVersion configuration
    variable is not set, or if we don't know the magic ABI values for the
    python version (no matter from which source the version came from).
    The second value will be None if a python version could not be resolved
    from the given pathname.
    """

    ver_from_path = None
    m = python_regex.search(path)
    if m:
        ver_from_path = m.group(1)

    expected_version = ver_from_path or python_default_version
    expected_magic_values = _python_magic_values.get(expected_version)

    if not expected_magic_values:
        return (None, ver_from_path)

    # In Python 2, if Py_UnicodeFlag is set, Python's import code uses a value
    # one higher, but this is off by default. In Python 3.0 and 3.1 (but no
    # longer in 3.2), it always uses the value one higher:
    if expected_version[:3] in ('3.0', '3.1'):
        expected_magic_values = [x + 1 for x in expected_magic_values]

    return (expected_magic_values, ver_from_path)


def py_demarshal_long(b):
    """
    Counterpart to Python's PyMarshal_ReadLongFromFile, operating on the
    bytes in a string.
    """
    if isinstance(b, str):
        b = map(ord, b)
    return (b[0] + (b[1] << 8) + (b[2] << 16) + (b[3] << 24))


def pyc_magic_from_chunk(chunk):
    """From given chunk (beginning of the file), return Python magic number"""
    return py_demarshal_long(chunk[:4]) & 0xffff


def pyc_mtime_from_chunk(chunk):
    """From given chunk (beginning of the file), return mtime or None

    From Python 3.7, mtime is not always present.

    See https://www.python.org/dev/peps/pep-0552/#specification
    """
    magic = pyc_magic_from_chunk(chunk)
    second = py_demarshal_long(chunk[4:8])
    if magic >= _python_magic_values['3.7'][0]:
        if second == 0:
            return py_demarshal_long(chunk[8:12])
        return None  # No mtime saved, TODO check hashes instead
    return second


def python_bytecode_to_script(path):
    """
    Given a python bytecode path, give the path of the .py file
    (or None if not python bytecode).
    """

    res = python_bytecode_regex_pep3147.search(path)
    if res:
        return res.group(1) + '/' + res.group(2) + '.py'

    res = python_bytecode_regex.search(path)
    if res:
        return res.group(1) + '.py'

    return None


def script_interpreter(chunk):
    res = shebang_regex.search(chunk) if chunk else None
    return (byte_to_string(res.group(1)), byte_to_string(res.group(2)).strip()) \
        if res and res.start() == 0 else (None, '')


def find_perm_in_tmpfiles(pkg, fname):
    """
    Look for file path in all tmpdfiles.d file declared in this pkg and return
    the permission column.
    """

    tmpd = []
    perms = '0644'
    user = 'root'
    group = 'root'
    fname = os.path.realpath(fname)

    for k, v in pkg.files.items():
        if 'tmpfiles.d' not in k or not k.endswith('.conf'):
            continue
        if not os.path.exists(v.path) or os.path.isdir(v.path):
            continue

        with open(v.path) as f:
            try:
                tmpd += f.readlines()
            except ValueError:
                # Can't read this file, so we are not trying to read definition
                # from there
                pass

    for line in tmpd:
        if f' {fname} ' not in line:
            continue
        try:
            _t, _p, perms, user, group, *_rest = line.split()
        except IndexError:
            continue

    return perms, user, group


class FilesCheck(AbstractCheck):
    man_regex = re.compile(r'/man(?:\d[px]?|n)/')
    info_regex = re.compile(r'(/usr/share|/usr)/info/')

    def __init__(self, config, output):
        super().__init__(config, output)
        self.use_debugsource = self.config.configuration['UseDebugSource']
        self.games_group_regex = re.compile(self.config.configuration['RpmGamesGroup'])
        self.dangling_exceptions = self.config.configuration['DanglingSymlinkExceptions']
        for item in self.dangling_exceptions.values():
            item['path'] = re.compile(item['path'])
        self.module_rpms_ok = self.config.configuration['KernelModuleRPMsOK']
        self.python_default_version = self.config.configuration['PythonDefaultVersion']
        self.perl_version_trick = self.config.configuration['PerlVersionTrick']
        self.skipdocs_regex = re.compile(self.config.configuration['SkipDocsRegexp'], re.IGNORECASE)
        self.meta_package_regex = re.compile(self.config.configuration['MetaPackageRegexp'])
        self.use_relative_symlinks = self.config.configuration['UseRelativeSymlinks']
        self.standard_groups = self.config.configuration['StandardGroups']
        self.standard_users = self.config.configuration['StandardUsers']
        self.disallowed_dirs = self.config.configuration['DisallowedDirs']
        self.compress_ext = self.config.configuration['CompressExtension']
        self.output.error_details.update({
            'non-standard-uid':
            """A file in this package is owned by a non standard user.
            Standard users are:
            %s.""" % ', '.join(self.standard_users),

            'non-standard-gid':
            """A file in this package is owned by a non standard group.
            Standard groups are:
            %s.""" % ', '.join(self.standard_groups),

            'manpage-not-compressed':
                """This manual page is not compressed with the %s compression method
                (does not have the %s extension). If the compression does not happen
                automatically when the package is rebuilt, make sure that you have the
                appropriate rpm helper and/or config packages for your target distribution
                installed and try rebuilding again; if it still does not happen automatically,
                you can compress this file in the %%install section of the spec file."""
                % (self.compress_ext, self.compress_ext),

            'infopage-not-compressed':
                """This info page is not compressed with the %s compression method
                (does not have the %s extension). If the compression does not happen
                automatically when the package is rebuilt, make sure that you have the
                appropriate rpm helper and/or config packages for your target distribution
                installed and try rebuilding again; if it still does not happen automatically,
                you can compress this file in the %%install section of the spec file."""
                % (self.compress_ext, self.compress_ext),
        })
        for i in self.disallowed_dirs:
            self.output.error_details.update({'dir-or-file-in-%s' % '-'.join(i.split('/')[1:]):
                                             """A file in the package is located in %s. It's not permitted
        for packages to install files in this directory.""" % i})

    def peek(self, filename, pkg, length=2048):
        """
        Peek into a file, return a chunk from its beginning and a flag if it
        seems to be a text file.
        """
        chunk = None
        try:
            with open(filename, 'rb') as fobj:
                chunk = fobj.read(length)
        except OSError as e:  # eg. https://bugzilla.redhat.com/209876
            self.output.add_info('W', pkg, 'read-error', e)
            return (chunk, False)

        if b'\0' in chunk:
            return (chunk, False)

        if not chunk:  # Empty files are considered text
            return (chunk, True)

        fl = filename.lower()

        # PDF's are binary but often detected as text by the algorithm below
        if fl.endswith('.pdf') and chunk.startswith(b'%PDF-'):
            return (chunk, False)
        # Ditto RDoc RI files
        if fl.endswith('.ri') and '/ri/' in fl:
            return (chunk, False)
        # And Sphinx inventory files
        if fl.endswith('.inv') and chunk.startswith(b'# Sphinx inventory'):
            return (chunk, False)

        # Binary if control chars are > 30% of the string
        control_chars = chunk.translate(None, printable_extended_ascii)
        nontext_ratio = float(len(control_chars)) / float(len(chunk))
        istext = nontext_ratio <= 0.30

        return (chunk, istext)

    def check(self, pkg):
        self._check_utf8(pkg)

        # Rest of the checks are for binary packages only
        if pkg.is_source:
            return

        self.devel_pkg = False
        self.deps = pkg.requires + pkg.prereq
        self.config_files = pkg.config_files
        self.ghost_files = pkg.ghost_files
        self.req_names = pkg.req_names
        self.lib_package = lib_package_regex.search(pkg.name)
        self.is_kernel_package = kernel_package_regex.search(pkg.name)
        self.debuginfo_package = debuginfo_package_regex.search(pkg.name)
        self.debugsource_package = debugsource_package_regex.search(pkg.name)

        # report these errors only once
        self.perl_dep_error = False
        self.python_dep_error = False
        self.lib_file = False
        self.non_lib_file = None
        self.log_files = []
        self.logrotate_file = False
        self.debuginfo_srcs = False
        self.debuginfo_debugs = False

        # Prefetch scriptlets, strip quotes from them (#169)
        self.postin = pkg[rpm.RPMTAG_POSTIN] or \
            pkg.scriptprog(rpm.RPMTAG_POSTINPROG)
        if self.postin:
            self.postin = quotes_regex.sub('', self.postin)
        self.postun = pkg[rpm.RPMTAG_POSTUN] or \
            pkg.scriptprog(rpm.RPMTAG_POSTUNPROG)
        if self.postun:
            self.postun = quotes_regex.sub('', self.postun)

        # Unique (rdev, inode) combinations
        self.hardlinks = {}

        # All executable files from standard bin dirs (basename => [paths])
        # Hack: basenames with empty paths links are symlinks (not subject
        # to duplicate binary check, but yes for man page existence check)
        self.bindir_exes = {}

        # All man page 'base' names (without section etc extensions)
        self.man_basenames = set()

        self._check_devel(pkg)
        self._check_nodoc(pkg)
        self._check_meta_package(pkg)
        self._check_empty_debuginfo(pkg)

        # Iterate over all pkg.files
        self._check_files(pkg)

        self._check_log_files_without_logrotate(pkg)
        self._check_outside_libdir_files(pkg)
        self._check_debuginfo_without_sources(pkg)
        self._check_bindir_exes(pkg)

    def _check_utf8(self, pkg):
        for filename in pkg.header[rpm.RPMTAG_FILENAMES] or ():
            if not is_utf8_bytestr(filename):
                self.output.add_info('E', pkg, 'filename-not-utf8', byte_to_string(filename))

    def _check_devel(self, pkg):
        # Check if the package is a development package
        self.devel_pkg = devel_regex.search(pkg.name)

        if not self.devel_pkg:
            for p in pkg.provides:
                if devel_regex.search(p[0]):
                    self.devel_pkg = True
                    break

    def _check_nodoc(self, pkg):
        if not self.lib_package and not pkg.doc_files:
            self.output.add_info('W', pkg, 'no-documentation')

    def _check_meta_package(self, pkg):
        if pkg.files and self.meta_package_regex.search(pkg.name):
            self.output.add_info('W', pkg, 'file-in-meta-package')

    def _check_empty_debuginfo(self, pkg):
        debuginfo = self.debuginfo_package or self.debugsource_package
        if not pkg.files and debuginfo:
            self.output.add_info('E', pkg, 'empty-debuginfo-package')

    def _check_log_files_without_logrotate(self, pkg):
        if len(self.log_files) and not self.logrotate_file:
            self.output.add_info('W', pkg, 'log-files-without-logrotate', sorted(self.log_files))

    def _check_outside_libdir_files(self, pkg):
        if self.lib_package and self.lib_file and self.non_lib_file:
            self.output.add_info('E', pkg, 'outside-libdir-files', self.non_lib_file)

    def _check_debuginfo_without_sources(self, pkg):
        if not self.use_debugsource and self.debuginfo_package and self.debuginfo_debugs and not self.debuginfo_srcs:
            self.output.add_info('E', pkg, 'debuginfo-without-sources')

    def _check_bindir_exes(self, pkg):
        for exe, paths in self.bindir_exes.items():
            if len(paths) > 1:
                self.output.add_info('W', pkg, 'duplicate-executable', exe, paths)
            if exe not in self.man_basenames:
                self.output.add_info('W', pkg, 'no-manual-page-for-binary', exe)

    def _check_files(self, pkg):
        for f, pkgfile in pkg.files.items():
            self._check_file(pkg, f, pkgfile)

    def _check_file(self, pkg, fname, pkgfile):
        if log_regex.search(fname):
            self.log_files.append(fname)

        self._check_file_manpage(pkg, fname)
        self._check_file_infopage_compressed(pkg, fname)
        self._check_file_unexpandaed_macro(pkg, fname)
        self._check_file_non_standard_uid(pkg, fname, pkgfile)
        self._check_file_non_standard_gid(pkg, fname, pkgfile)
        self._check_file_kernel_modules(pkg, fname)
        self._check_file_dir_or_file(pkg, fname)
        self._check_file_non_ghost_in_run(pkg, fname)
        self._check_file_systemd_unit_in_etc(pkg, fname)
        self._check_file_udev_rule_in_etc(pkg, fname)
        self._check_file_tmpfiles_conf_in_etc(pkg, fname)
        self._check_file_subdir_in_bin(pkg, fname)
        self._check_file_siteperl_in_perl_module(pkg, fname)
        self._check_file_backup_file_in_package(pkg, fname)
        self._check_file_version_control_internal_file(pkg, fname)
        self._check_file_htaccess_file(pkg, fname)
        self._check_file_hidden_file_or_dir(pkg, fname)
        self._check_file_manifest_in_perl_module(pkg, fname)
        self._check_file_info_dir_file(pkg, fname)
        self._check_file_makefile_junk(pkg, fname)
        self._check_file_logrotate(pkg, fname)
        self._check_file_crontab(pkg, fname)
        self._check_file_xinetd(pkg, fname)
        self._check_file_compressed_symlink(pkg, fname, pkgfile)
        self._check_file_hardlink(pkg, fname, pkgfile)

        # normal file check
        self._check_file_normal_file(pkg, fname, pkgfile)
        # normal dir check
        self._check_file_dir(pkg, fname, pkgfile)
        # symbolic link check
        self._check_file_link(pkg, fname, pkgfile)

        self._check_file_crond(pkg, fname, pkgfile)

        # files with all permissions bits as zero
        # https://github.com/rpm-software-management/rpmlint/issues/878
        self._check_file_zero_perms(pkg, fname, pkgfile)

    def _check_file_manpage(self, pkg, fname):
        """
        Check if the the manual page is compressed with the compression method
        stated in the rpmlint configuration (CompressExtension option).
        Check also for a correct manual page location and if not included in a subfolder.

        Print a warning if it's not compressed.
        """
        if stat.S_ISDIR(pkg.files[fname].mode):
            return

        res = man_base_regex.fullmatch(fname)
        if not res:
            return

        category = res.group('category')
        filename = Path(res.group('filename'))
        suffixes = filename.suffixes

        if self.compress_ext:
            if self.compress_ext != suffixes[-1][1:]:
                self.output.add_info('W', pkg, 'manpage-not-compressed',
                                     self.compress_ext, fname)
            suffixes = suffixes[:-1]

        if suffixes:
            file_category = suffixes[-1][1:]
            if not file_category.startswith(category):
                self.output.add_info('E', pkg, 'bad-manual-page-folder', fname, f'expected folder: man{file_category}')
            if str(filename.parent) != '.':
                self.output.add_info('E', pkg, 'manual-page-in-subfolder', fname)

    def _check_file_infopage_compressed(self, pkg, fname):
        """
        Check if the the info page is compressed with the compression method
        stated in the rpmlint configuration (CompressExtension option).

        Print a warning if it's not compressed.
        """
        if self.compress_ext and self.info_regex.search(fname) and \
                not fname.endswith('/info/dir') and not fname.endswith(self.compress_ext):
            self.output.add_info('W', pkg, 'infopage-not-compressed',
                                 self.compress_ext, fname)

    def _check_file_crond(self, pkg, fname, pkgfile):
        if not fname.startswith('/etc/cron.d/'):
            return

        mode = pkgfile.mode
        mode_is_exec = mode & 0o111
        if stat.S_ISLNK(mode):
            self.output.add_info('E', pkg, 'symlink-crontab-file', fname)

        if mode_is_exec:
            self.output.add_info('E', pkg, 'executable-crontab-file', fname)

        if stat.S_IWGRP & mode or stat.S_IWOTH & mode:
            self.output.add_info('E', pkg, 'non-owner-writeable-only-crontab-file', fname)

    def _check_file_zero_perms(self, pkg, fname, pkgfile):
        mode = pkgfile.mode
        perm = mode & 0o7777
        if not perm:
            if pkgfile.is_ghost:
                perms, user, group = find_perm_in_tmpfiles(pkg, pkgfile.name)
                suggestion = f'Suggestion: "%ghost %attr({perms},{user},{group}) {pkgfile.name}"'
                # Suggest the attrs if there's a tmpfiles.d
                self.output.add_info('W', pkg, 'zero-perms-ghost', suggestion)
            else:
                self.output.add_info('W', pkg, 'zero-perms', fname, '%o' % perm)

    def _check_file_unexpandaed_macro(self, pkg, fname):
        for match in self.macro_regex.findall(fname):
            self.output.add_info('W', pkg, 'unexpanded-macro', fname, match)

    def _check_file_non_standard_uid(self, pkg, fname, pkgfile):
        if pkgfile.user not in self.standard_users:
            self.output.add_info('W', pkg, 'non-standard-uid', fname, pkgfile.user)

    def _check_file_non_standard_gid(self, pkg, fname, pkgfile):
        if pkgfile.group not in self.standard_groups:
            self.output.add_info('W', pkg, 'non-standard-gid', fname, pkgfile.group)

    def _check_file_kernel_modules(self, pkg, fname):
        if not self.module_rpms_ok and kernel_modules_regex.search(fname) and not \
                self.is_kernel_package:
            self.output.add_info('E', pkg, 'kernel-modules-not-in-kernel-packages', fname)

    def _check_file_dir_or_file(self, pkg, fname):
        for i in self.disallowed_dirs:
            if fname.startswith(i):
                self.output.add_info('E', pkg, 'dir-or-file-in-%s' %
                                     '-'.join(i.split('/')[1:]), fname)

    def _check_file_non_ghost_in_run(self, pkg, fname):
        if fname.startswith('/run/'):
            if fname not in self.ghost_files:
                self.output.add_info('W', pkg, 'non-ghost-in-run', fname)

    def _check_file_systemd_unit_in_etc(self, pkg, fname):
        if fname.startswith('/etc/systemd/system/'):
            self.output.add_info('W', pkg, 'systemd-unit-in-etc', fname)

    def _check_file_udev_rule_in_etc(self, pkg, fname):
        if fname.startswith('/etc/udev/rules.d/'):
            self.output.add_info('W', pkg, 'udev-rule-in-etc', fname)

    def _check_file_tmpfiles_conf_in_etc(self, pkg, fname):
        if fname.startswith('/etc/tmpfiles.d/'):
            self.output.add_info('W', pkg, 'tmpfiles-conf-in-etc', fname)

    def _check_file_subdir_in_bin(self, pkg, fname):
        if sub_bin_regex.search(fname):
            self.output.add_info('E', pkg, 'subdir-in-bin', fname)

    def _check_file_siteperl_in_perl_module(self, pkg, fname):
        if '/site_perl/' in fname:
            self.output.add_info('W', pkg, 'siteperl-in-perl-module', fname)

    def _check_file_backup_file_in_package(self, pkg, fname):
        if backup_regex.search(fname):
            self.output.add_info('E', pkg, 'backup-file-in-package', fname)

    def _check_file_version_control_internal_file(self, pkg, fname):
        if scm_regex.search(fname):
            self.output.add_info('E', pkg, 'version-control-internal-file', fname)

    def _check_file_htaccess_file(self, pkg, fname):
        if fname.endswith('/.htaccess'):
            self.output.add_info('E', pkg, 'htaccess-file', fname)

    def _check_file_hidden_file_or_dir(self, pkg, fname):
        if (hidden_file_regex.search(fname) and
                not fname.startswith('/etc/skel/') and
                not fname.endswith('/.build-id') and
                not fname.endswith('/.cargo-checksum.json')):
            self.output.add_info('W', pkg, 'hidden-file-or-dir', fname)

    def _check_file_manifest_in_perl_module(self, pkg, fname):
        if manifest_perl_regex.search(fname):
            self.output.add_info('W', pkg, 'manifest-in-perl-module', fname)

    def _check_file_info_dir_file(self, pkg, fname):
        if fname == '/usr/info/dir' or fname == '/usr/share/info/dir':
            self.output.add_info('E', pkg, 'info-dir-file', fname)

    def _check_file_makefile_junk(self, pkg, fname):
        fpath = Path(fname)
        is_doc = fname in pkg.doc_files
        if fpath.name == 'Makefile.am' and str(fpath.with_suffix('.in')) in pkg.files and is_doc:
            self.output.add_info('W', pkg, 'makefile-junk', fname)

    def _check_file_logrotate(self, pkg, fname):
        res = logrotate_regex.search(fname)
        if res:
            self.logrotate_file = True
            if res.group(1) != pkg.name:
                self.output.add_info('E', pkg, 'incoherent-logrotate-file', fname)

        deps = [x[0] for x in pkg.requires + pkg.recommends + pkg.suggests]
        if res and 'logrotate' not in deps and pkg.name != 'logrotate':
            self.output.add_info('E', pkg, 'missing-dependency-to-logrotate', 'for logrotate script', fname)

    def _check_file_crontab(self, pkg, fname):
        deps = [x[0] for x in pkg.requires + pkg.recommends + pkg.suggests]
        if fname.startswith('/etc/cron.') and 'crontabs' not in deps and pkg.name != 'crontabs':
            self.output.add_info('E', pkg, 'missing-dependency-to-crontabs', 'for cron script', fname)

    def _check_file_xinetd(self, pkg, fname):
        deps = [x[0] for x in pkg.requires + pkg.recommends + pkg.suggests]
        if fname.startswith('/etc/xinet.d/') and 'xinetd' not in deps and pkg.name != 'xinetd':
            self.output.add_info('E', pkg, 'missing-dependency-to-xinetd', 'for xinet.d script', fname)

    def _check_file_compressed_symlink(self, pkg, fname, pkgfile):
        link = pkgfile.linkto
        if link != '':
            ext = compr_regex.search(link)
            if ext and not re.compile(r'\.%s$' % ext.group(1)).search(fname):
                self.output.add_info('E', pkg, 'compressed-symlink-with-wrong-ext',
                                     fname, link)

    def _check_file_hardlink(self, pkg, fname, pkgfile):
        rdev = pkgfile.rdev
        inode = pkgfile.inode

        # Hardlink check
        for hardlink in self.hardlinks.get((rdev, inode), ()):
            if Path(hardlink).parent != Path(fname).parent:
                self.output.add_info('W', pkg, 'cross-directory-hard-link', fname, hardlink)
        self.hardlinks.setdefault((rdev, inode), []).append(fname)

    def _check_file_link_devel(self, pkg, fname, pkgfile):
        is_so = sofile_regex.search(fname)
        if not self.devel_pkg and is_so and not pkgfile.linkto.endswith('.so'):
            self.output.add_info('W', pkg, 'devel-file-in-non-devel-package', fname)

    def _check_file_link_man(self, pkg, fname):
        res = man_base_regex.fullmatch(fname)
        if res:
            self.man_basenames.add(res.group('binary'))

    def _check_file_link_bindir_exes(self, pkg, fname):
        res = bin_regex.search(fname)
        if res:
            exe = res.group(1)
            if '/' not in exe:
                self.bindir_exes.setdefault(exe, [])

    def _check_file_link_absolute(self, pkg, fname, pkgfile):
        link = pkgfile.linkto

        # absolute link
        r = absolute_regex.search(link)
        if not r:
            return

        is_so = sofile_regex.search(fname)
        if not is_so and link not in pkg.files and \
                link not in self.req_names:
            is_exception = False
            for e in self.dangling_exceptions.values():
                if e['path'].search(link):
                    is_exception = e['name']
                    break
            if is_exception:
                if is_exception not in self.req_names:
                    self.output.add_info('W', pkg, 'no-dependency-on',
                                         is_exception)
            else:
                self.output.add_info('W', pkg, 'dangling-symlink', fname, link)
        linktop = r.group(1)

        r = absolute_regex.search(fname)
        if r:
            filetop = r.group(1)
            if filetop == linktop or self.use_relative_symlinks:
                self.output.add_info('W', pkg, 'symlink-should-be-relative',
                                     fname, link)

    def _check_file_link_relative(self, pkg, fname, pkgfile):
        link = pkgfile.linkto

        # relative link
        r = absolute_regex.search(link)
        if r:
            return

        is_so = sofile_regex.search(fname)
        if not is_so:
            abslink = f'{Path(fname).parent}/{link}'
            abslink = os.path.normpath(abslink)
            if abslink not in pkg.files and abslink not in self.req_names:
                is_exception = False
                for e in self.dangling_exceptions.values():
                    if e['path'].search(link):
                        is_exception = e['name']
                        break
                if is_exception:
                    if is_exception not in self.req_names:
                        self.output.add_info('W', pkg, 'no-dependency-on',
                                             is_exception)
                else:
                    self.output.add_info('W', pkg, 'dangling-relative-symlink',
                                         fname, link)
        pathcomponents = fname.split('/')[1:]
        r = points_regex.search(link)
        lastpop = None
        mylink = None

        while r:
            mylink = r.group(1)
            if len(pathcomponents) == 0:
                self.output.add_info('E', pkg, 'symlink-has-too-many-up-segments',
                                     fname, link)
                break
            else:
                lastpop = pathcomponents[0]
                pathcomponents = pathcomponents[1:]
                r = points_regex.search(mylink)

        if mylink and lastpop:
            r = absolute2_regex.search(mylink)
            linktop = r.group(1)

            # does the link go up and then down into the same
            # directory?
            # if linktop == lastpop:
            #    self.output.add_info('W', pkg, 'lengthy-symlink', f, link)

            # have we reached the root directory?
            if len(pathcomponents) == 0 and linktop != lastpop \
                    and not self.use_relative_symlinks:
                # relative link into other toplevel directory
                self.output.add_info('W', pkg, 'symlink-should-be-absolute', fname,
                                     link)
            # check additional segments for mistakes like
            # `foo/../bar/'
            for linksegment in mylink.split('/'):
                if linksegment == '..':
                    self.output.add_info('E',
                                         pkg,
                                         'symlink-contains-up-and-down-segments',
                                         fname, link)

    def _check_file_link_bindir_shebang(self, pkg, fname, pkgfile):
        basedir = Path(fname).parent
        linkto = str((basedir / Path(pkgfile.linkto)).resolve())
        # Link to a file not in the package, so ignore
        if linkto not in pkg.files:
            return

        realbin = pkg.files[linkto]
        # Link to something in bindir is okay
        if bin_regex.search(realbin.name):
            return
        if not stat.S_ISREG(realbin.mode):
            return

        file_chunk, file_istext = self.peek(realbin.path, pkg)
        file_interpreter, _file_interpreter_args = script_interpreter(file_chunk)
        # Not a script with shebang, so ignore
        if not file_interpreter:
            return

        # If the shebang interpreter is a dependency, it's okay
        deps = [x[0] for x in pkg.requires]
        if file_interpreter in deps:
            return

        self.output.add_info('W', pkg, 'symlink-to-binary-with-shebang', fname,
                             f'is a link to a script ({realbin.name}) but missing'
                             f' requires for {file_interpreter}')

    def _check_file_link(self, pkg, fname, pkgfile):
        if not stat.S_ISLNK(pkgfile.mode):
            return

        self._check_file_link_devel(pkg, fname, pkgfile)
        self._check_file_link_man(pkg, fname)
        self._check_file_link_bindir_exes(pkg, fname)
        self._check_file_link_absolute(pkg, fname, pkgfile)
        self._check_file_link_relative(pkg, fname, pkgfile)
        self._check_file_link_bindir_shebang(pkg, fname, pkgfile)

    def _check_file_dir(self, pkg, fname, pkgfile):
        if not stat.S_ISDIR(pkgfile.mode):
            return

        mode = pkgfile.mode
        perm = mode & 0o7777
        if mode & 0o1002 == 2:  # world writable w/o sticky bit
            self.output.add_info('E', pkg, 'world-writable', fname, '%o' % perm)
        if perm != 0o755:
            self.output.add_info('E', pkg, 'non-standard-dir-perm', fname, '%o' % perm)
        if pkg.name not in filesys_packages and fname in STANDARD_DIRS:
            self.output.add_info('E', pkg, 'standard-dir-owned-by-package', fname)
        if hidden_file_regex.search(fname) and not fname.endswith('/.build-id'):
            self.output.add_info('W', pkg, 'hidden-file-or-dir', fname)

    def _check_file_normal_file(self, pkg, fname, pkgfile):
        if not stat.S_ISREG(pkgfile.mode):
            return

        self._file_nonexec_file = False
        self._file_chunk = None
        self._file_istext = False
        self._file_interpreter = None
        self._file_interpreter_args = None
        self._file_is_buildconfig = False

        # set[ug]id bit check
        self._check_file_normal_file_setuid_bit(pkg, fname, pkgfile)
        self._check_file_normal_file_libfile(pkg, fname)
        self._check_file_normal_file_logfile(pkg, fname, pkgfile)
        # Fill class attributes, chunk, istext, interpreter, is_buildconfig
        self._check_file_normal_file_getdata(pkg, fname, pkgfile)
        self._check_file_normal_file_doc(pkg, fname)
        self._check_file_normal_file_non_devel(pkg, fname)
        self._check_file_normal_file_lib(pkg, fname, pkgfile)
        self._check_file_normal_file_depmod_call(pkg, fname)
        self._check_file_normal_file_install_info(pkg, fname)
        self._check_file_normal_file_perl_temp(pkg, fname)
        self._check_file_normal_file_rpaths_in_buildconfig(pkg, fname)
        # look for man pages
        self._check_file_normal_file_look_for_man(pkg, fname)
        self._check_file_normal_file_bin(pkg, fname, pkgfile)
        self._check_file_normal_file_devel(pkg, fname, pkgfile)
        self._check_file_normal_file_non_readable(pkg, fname, pkgfile)
        self._check_file_normal_file_zero_length(pkg, fname, pkgfile)
        self._check_file_normal_file_world_w(pkg, fname, pkgfile)
        self._check_file_normal_file_perl_dep(pkg, fname)
        self._check_file_normal_file_python_dep(pkg, fname)
        self._check_file_normal_file_python_source(pkg, fname)
        # normal executable check
        self._check_file_normal_file_exec(pkg, fname, pkgfile)
        self._check_file_normal_file_non_conf_in_etc(pkg, fname)
        self._check_file_normal_file_python_noarch(pkg, fname)
        self._check_file_normal_file_debuginfo(pkg, fname)
        self._check_file_normal_file_gzipped_svg(pkg, fname)
        self._check_file_normal_file_pem(pkg, fname)
        self._check_file_normal_file_tcl(pkg, fname)
        # text file checks
        self._check_file_normal_file_text(pkg, fname, pkgfile)
        self._check_file_normal_file_not_utf8(pkg, fname, pkgfile)

    def _check_file_normal_file_setuid_bit(self, pkg, fname, pkgfile):
        user = pkgfile.user
        group = pkgfile.group
        mode = pkgfile.mode
        perm = mode & 0o7777
        if stat.S_ISGID & mode or stat.S_ISUID & mode:
            if stat.S_ISUID & mode:
                self.output.add_info('E', pkg, 'setuid-binary', fname, user, '%o' % perm)
            if (stat.S_ISGID & mode and
                    not (group == 'games' and
                         (games_path_regex.search(fname) or
                          self.games_group_regex.search(pkg[rpm.RPMTAG_GROUP])))):
                self.output.add_info('E', pkg, 'setgid-binary', fname, group,
                                     '%o' % perm)
            if mode & 0o777 != 0o755:
                self.output.add_info('E', pkg, 'non-standard-executable-perm', fname,
                                     '%o' % perm)

    def _check_file_normal_file_libfile(self, pkg, fname):
        is_doc = fname in pkg.doc_files

        if not self.devel_pkg:
            if lib_path_regex.search(fname):
                self.lib_file = True
            elif not is_doc:
                self.non_lib_file = fname

    def _check_file_normal_file_logfile(self, pkg, fname, pkgfile):
        user = pkgfile.user
        group = pkgfile.group

        if log_regex.search(fname):
            self._file_nonexec_file = True
            if user != 'root':
                self.output.add_info('E', pkg, 'non-root-user-log-file', fname, user)
            if group != 'root':
                self.output.add_info('E', pkg, 'non-root-group-log-file', fname, group)
            if fname not in self.ghost_files:
                self.output.add_info('E', pkg, 'non-ghost-file', fname)

    def _check_file_normal_file_getdata(self, pkg, fname, pkgfile):
        res = None
        try:
            res = os.access(pkgfile.path, os.R_OK)
        except UnicodeError as e:  # e.g. non-ASCII, C locale, python 3
            self.output.add_info('W', pkg, 'inaccessible-filename', fname, e)
        else:
            if res:
                (self._file_chunk, self._file_istext) = self.peek(pkgfile.path, pkg)

        (self._file_interpreter, self._file_interpreter_args) = script_interpreter(self._file_chunk)
        self._file_is_buildconfig = self._file_istext and buildconfigfile_regex.search(fname)

    def _check_file_normal_file_doc(self, pkg, fname):
        is_doc = fname in pkg.doc_files
        if doc_regex.search(fname):
            if not self._file_interpreter:
                self._file_nonexec_file = True
            if not is_doc:
                self.output.add_info('E', pkg, 'not-listed-as-documentation', fname)

    def _check_file_normal_file_non_devel(self, pkg, fname):
        if self.devel_pkg and fname.endswith('.typelib'):
            self.output.add_info('E', pkg, 'non-devel-file-in-devel-package', fname)

    def _check_file_normal_file_lib(self, pkg, fname, pkgfile):
        # check ldconfig call in %post and %postun
        mode = pkgfile.mode
        if lib_regex.search(fname):
            if self.devel_pkg and not (sofile_regex.search(fname) and stat.S_ISLNK(mode)):
                self.output.add_info('E', pkg, 'non-devel-file-in-devel-package', fname)
            if not self.postin:
                self.output.add_info('E', pkg, 'library-without-ldconfig-postin', fname)
            else:
                if not ldconfig_regex.search(self.postin):
                    self.output.add_info('E', pkg, 'postin-without-ldconfig', fname)

            if not self.postun:
                self.output.add_info('E', pkg, 'library-without-ldconfig-postun', fname)
            else:
                if not ldconfig_regex.search(self.postun):
                    self.output.add_info('E', pkg, 'postun-without-ldconfig', fname)

    def _check_file_normal_file_depmod_call(self, pkg, fname):
        # check depmod call in %post and %postun
        res = not self.is_kernel_package and kernel_modules_regex.search(fname)
        if res:
            kernel_version = res.group(1)
            kernel_version_regex = re.compile(
                r'\bdepmod\s+-a.*F\s+/boot/System\.map-' +
                re.escape(kernel_version) + r'\b.*\b' +
                re.escape(kernel_version) + r'\b',
                re.MULTILINE | re.DOTALL)

            if not self.postin or not depmod_regex.search(self.postin):
                self.output.add_info('E', pkg, 'module-without-depmod-postin', fname)
            # check that we run depmod on the right kernel
            elif not kernel_version_regex.search(self.postin):
                self.output.add_info('E', pkg, 'postin-with-wrong-depmod', fname)

            if not self.postun or not depmod_regex.search(self.postun):
                self.output.add_info('E', pkg, 'module-without-depmod-postun', fname)
            # check that we run depmod on the right kernel
            elif not kernel_version_regex.search(self.postun):
                self.output.add_info('E', pkg, 'postun-with-wrong-depmod', fname)

    def _check_file_normal_file_install_info(self, pkg, fname):
        # check install-info call in %post and %postun
        if fname.startswith('/usr/share/info/'):
            if not self.postin:
                self.output.add_info('E', pkg,
                                     'info-files-without-install-info-postin', fname)
            elif not install_info_regex.search(self.postin):
                self.output.add_info('E', pkg, 'postin-without-install-info', fname)

            preun = pkg[rpm.RPMTAG_PREUN] or \
                pkg.scriptprog(rpm.RPMTAG_PREUNPROG)
            if not self.postun and not preun:
                self.output.add_info('E', pkg,
                                     'info-files-without-install-info-postun', fname)
            elif not ((self.postun and install_info_regex.search(self.postun)) or
                      (preun and install_info_regex.search(preun))):
                self.output.add_info('E', pkg, 'postin-without-install-info', fname)

    def _check_file_normal_file_perl_temp(self, pkg, fname):
        # check perl temp file
        if perl_temp_file_regex.search(fname):
            self.output.add_info('W', pkg, 'perl-temp-file', fname)

    def _check_file_normal_file_rpaths_in_buildconfig(self, pkg, fname):
        # check rpaths in buildconfig files
        if self._file_is_buildconfig:
            ln = pkg.grep(buildconfig_rpath_regex, fname)
            if ln:
                self.output.add_info('E', pkg, 'rpath-in-buildconfig', fname, 'lines', ln)

    def _check_file_normal_file_look_for_man(self, pkg, fname):
        res = man_base_regex.fullmatch(fname)
        if res:
            self.man_basenames.add(res.group('binary'))

    def _check_file_normal_file_bin(self, pkg, fname, pkgfile):
        mode = pkgfile.mode
        perm = mode & 0o7777
        mode_is_exec = mode & 0o111
        res = bin_regex.search(fname)
        if res:
            if not mode_is_exec:
                self.output.add_info('W', pkg, 'non-executable-in-bin', fname,
                                     '%o' % perm)
            else:
                exe = res.group(1)
                if '/' not in exe:
                    self.bindir_exes.setdefault(exe, []).append(fname)

    def _check_file_normal_file_devel(self, pkg, fname, pkgfile):
        is_doc = fname in pkg.doc_files
        if (not self.devel_pkg and not is_doc and
            (self._file_is_buildconfig or includefile_regex.search(fname) or
             develfile_regex.search(fname))):
            self.output.add_info('W', pkg, 'devel-file-in-non-devel-package', fname)

    def _check_file_normal_file_non_readable(self, pkg, fname, pkgfile):
        # Do not check permissions for ghosts files
        # https://github.com/rpm-software-management/rpmlint/issues/1287
        if pkgfile.is_ghost:
            return

        mode = pkgfile.mode
        perm = mode & 0o7777
        if mode & 0o444 != 0o444 and perm & 0o7000 == 0:
            ok_nonreadable = False
            for regex in non_readable_regexs:
                if regex.search(fname):
                    ok_nonreadable = True
                    break
            if not ok_nonreadable:
                self.output.add_info('E', pkg, 'non-readable', fname, '%o' % perm)

    def _check_file_normal_file_zero_length(self, pkg, fname, pkgfile):
        size = pkgfile.size
        if size == 0 and not normal_zero_length_regex.search(fname) and \
                fname not in self.ghost_files:
            self.output.add_info('E', pkg, 'zero-length', fname)

    def _check_file_normal_file_world_w(self, pkg, fname, pkgfile):
        mode = pkgfile.mode
        perm = mode & 0o7777
        if mode & stat.S_IWOTH:
            self.output.add_info('E', pkg, 'world-writable', fname, '%o' % perm)

    def _check_file_normal_file_perl_dep(self, pkg, fname):
        if not self.perl_dep_error:
            res = perl_regex.search(fname)
            if res:
                deps = [x[0] for x in pkg.requires + pkg.recommends + pkg.suggests]
                if self.perl_version_trick:
                    vers = res.group(1) + '.' + res.group(2)
                else:
                    vers = res.group(1) + res.group(2)
                if not (pkg.check_versioned_dep('perl-base', vers) or
                        pkg.check_versioned_dep('perl', vers) or
                        f'perl(:MODULE_COMPAT_{vers})' in deps):
                    self.output.add_info('E', pkg, 'no-dependency-on',
                                         'perl-base', vers)
                    self.perl_dep_error = True

    def _check_file_normal_file_python_dep(self, pkg, fname):
        if not self.python_dep_error:
            res = python_regex.search(fname)
            if (res and not
                    any(pkg.check_versioned_dep(dep, res.group(1))
                        for dep in (
                            'python', 'python-base', 'python(abi)'))):
                self.output.add_info('E', pkg, 'no-dependency-on', 'python-base',
                                     res.group(1))
                self.python_dep_error = True

    def _check_file_normal_file_python_source(self, pkg, fname):
        source_file = python_bytecode_to_script(fname)
        if not source_file:
            return

        if source_file in pkg.files:
            if self._file_chunk:
                # Verify that the magic ABI value embedded in the
                # .pyc header is correct
                found_magic = pyc_magic_from_chunk(self._file_chunk)
                exp_magic, exp_version = get_expected_pyc_magic(fname, self.python_default_version)
                if exp_magic and found_magic not in exp_magic:
                    found_version = 'unknown'
                    for (pv, pm) in _python_magic_values.items():
                        if found_magic in pm:
                            found_version = pv
                            break
                    # If expected version was from the file path,
                    # issue # an error, otherwise a warning.
                    msg = (pkg,
                           'python-bytecode-wrong-magic-value',
                           fname, 'expected %s (%s), found %d (%s)' %
                           (' or '.join(map(str, exp_magic)),
                            exp_version or self.python_default_version,
                            found_magic, found_version))
                    if exp_version is not None:
                        self.output.add_info('E', *msg)
                    else:
                        self.output.add_info('W', *msg)

                # Verify that the timestamp embedded in the .pyc
                # header matches the mtime of the .py file:
                pyc_timestamp = pyc_mtime_from_chunk(self._file_chunk)
                # If it's a symlink, check target file mtime.
                srcfile = pkg.readlink(pkg.files[source_file])
                if not srcfile:
                    self.output.add_info('W', pkg, 'python-bytecode-without-source', fname)
                elif (pyc_timestamp is not None and
                      pyc_timestamp != srcfile.mtime):
                    cts = datetime.fromtimestamp(
                        pyc_timestamp).isoformat()
                    sts = datetime.fromtimestamp(
                        srcfile.mtime).isoformat()
                    self.output.add_info('E',
                                         pkg, 'python-bytecode-inconsistent-mtime',
                                         fname, cts, srcfile.name, sts)
        else:
            self.output.add_info('W', pkg, 'python-bytecode-without-source', fname)

    def _check_file_normal_file_exec(self, pkg, fname, pkgfile):
        mode = pkgfile.mode
        perm = mode & 0o7777
        mode_is_exec = mode & 0o111
        if mode & stat.S_IXUSR and perm != 0o755:
            self.output.add_info('E', pkg, 'non-standard-executable-perm',
                                 fname, '%o' % perm)
        if mode_is_exec:
            if fname in self.config_files:
                self.output.add_info('E', pkg, 'executable-marked-as-config-file', fname)
            if not self._file_nonexec_file:
                # doc_regex and log_regex checked earlier, no match,
                # check rest of usual cases here.  Sourced scripts have
                # their own check, so disregard them here.
                self._file_nonexec_file = fname.endswith('.pc') or \
                    compr_regex.search(fname) or \
                    includefile_regex.search(fname) or \
                    develfile_regex.search(fname) or \
                    logrotate_regex.search(fname)
            if self._file_nonexec_file:
                self.output.add_info('W', pkg, 'spurious-executable-perm', fname)

    def _check_file_normal_file_non_conf_in_etc(self, pkg, fname):
        if fname.startswith('/etc/') and fname not in self.config_files and \
                fname not in self.ghost_files and not fname.startswith('/etc/ld.so.conf.d/'):
            self.output.add_info('W', pkg, 'non-conffile-in-etc', fname)

    def _check_file_normal_file_python_noarch(self, pkg, fname):
        if pkg.arch == 'noarch' and fname.startswith('/usr/lib64/python'):
            self.output.add_info('E', pkg, 'noarch-python-in-64bit-path', fname)

    def _check_file_normal_file_debuginfo(self, pkg, fname):
        if self.debuginfo_package:
            if fname.endswith('.debug'):
                self.debuginfo_debugs = True
            else:
                self.debuginfo_srcs = True

    def _check_file_normal_file_gzipped_svg(self, pkg, fname):
        if fname.endswith('.svgz') and fname[0:-1] not in pkg.files \
                and scalable_icon_regex.search(fname):
            self.output.add_info('W', pkg, 'gzipped-svg-icon', fname)

    def _check_file_normal_file_pem(self, pkg, fname):
        if fname.endswith('.pem') and fname not in self.ghost_files:
            if pkg.grep(start_certificate_regex, fname):
                self.output.add_info('W', pkg, 'pem-certificate', fname)
            if pkg.grep(start_private_key_regex, fname):
                self.output.add_info('E', pkg, 'pem-private-key', fname)

    def _check_file_normal_file_tcl(self, pkg, fname):
        if tcl_regex.search(fname):
            self.output.add_info('E', pkg, 'tcl-extension-file', fname)

    def _check_file_normal_file_text(self, pkg, fname, pkgfile):
        if not self._file_istext:
            return

        mode = pkgfile.mode
        perm = mode & 0o7777
        mode_is_exec = mode & 0o111
        is_doc = fname in pkg.doc_files

        # ignore perl module shebang -- TODO: disputed...
        if fname.endswith('.pm'):
            self._file_interpreter = None
        # sourced scripts should not be executable
        if sourced_script_regex.search(fname):
            if self._file_interpreter:
                self.output.add_info('E', pkg,
                                     'sourced-script-with-shebang', fname,
                                     self._file_interpreter, self._file_interpreter_args)
            if mode_is_exec:
                self.output.add_info('E', pkg, 'executable-sourced-script',
                                     fname, '%o' % perm)
        # ...but executed ones should
        elif self._file_interpreter or mode_is_exec or script_regex.search(fname):
            if self._file_interpreter:
                res = interpreter_regex.search(self._file_interpreter)
                if (mode_is_exec or script_regex.search(fname)):
                    if res and res.group(1) == 'env':
                        self.output.add_info('E', pkg, 'env-script-interpreter',
                                             fname, self._file_interpreter,
                                             self._file_interpreter_args)
                    elif not res:
                        self.output.add_info('E', pkg, 'wrong-script-interpreter',
                                             fname, self._file_interpreter,
                                             self._file_interpreter_args)
            elif not self._file_nonexec_file and not \
                    (lib_path_regex.search(fname) and
                     fname.endswith('.la')):
                self.output.add_info('E', pkg, 'script-without-shebang', fname)

            if not mode_is_exec and not is_doc and \
                    self._file_interpreter and self._file_interpreter.startswith('/'):
                self.output.add_info('E', pkg, 'non-executable-script', fname,
                                     '%o' % perm, self._file_interpreter,
                                     self._file_interpreter_args)
            if b'\r' in self._file_chunk:
                self.output.add_info('E', pkg, 'wrong-script-end-of-line-encoding', fname)
        elif is_doc and not self.skipdocs_regex.search(fname):
            if b'\r' in self._file_chunk:
                self.output.add_info('W', pkg, 'wrong-file-end-of-line-encoding', fname)
            # We check only doc text files for UTF-8-ness;
            # checking everything may be slow and can generate
            # lots of unwanted noise.
            if not is_utf8(pkgfile.path):
                self.output.add_info('W', pkg, 'file-not-utf8', fname)
        if fsf_license_regex.search(self._file_chunk) and \
                fsf_wrong_address_regex.search(self._file_chunk):
            self.output.add_info('E', pkg, 'incorrect-fsf-address', fname)

    def _check_file_normal_file_not_utf8(self, pkg, fname, pkgfile):
        is_doc = fname in pkg.doc_files
        if not self._file_istext and is_doc and self._file_chunk and compr_regex.search(fname):
            ff = compr_regex.sub('', fname)
            # compressed docs, eg. info and man files etc
            if not self.skipdocs_regex.search(ff) and not is_utf8(pkgfile.path):
                self.output.add_info('W', pkg, 'file-not-utf8', fname)



#############################################################################
# File          : I18NCheck.py
# Package       : rpmlint
# Author        : Frederic Lepied
# Created on    : Mon Nov 22 20:02:56 1999
# Purpose       : checks i18n bugs.
#############################################################################

import re

import rpm
from rpmlint.__isocodes__ import COUNTRIES, LANGUAGES
from rpmlint.checks.AbstractCheck import AbstractCheck


# Associative array of invalid value => correct value
INCORRECT_LOCALES = {
    'in': 'id',
    'in_ID': 'id_ID',
    'iw': 'he',
    'iw_IL': 'he_IL',
    'gr': 'el',
    'gr_GR': 'el_GR',
    'cz': 'cs',
    'cz_CZ': 'cs_CZ',
    'lug': 'lg',  # 'lug' is valid, but we standardize on 2 letter codes
    'en_UK': 'en_GB'}

package_regex = re.compile('-(' + '|'.join(LANGUAGES) + ')$')
locale_regex = re.compile('^(/usr/share/locale/([^/]+))/')
correct_subdir_regex = re.compile('^(([a-z][a-z]([a-z])?(_[A-Z][A-Z])?)([.@].*$)?)$')
lc_messages_regex = re.compile('/usr/share/locale/([^/]+)/LC_MESSAGES/.*(mo|po)$')
man_regex = re.compile('/usr(?:/share)?/man/([^/]+)/man[0-9n][^/]*/[^/]+$')

# list of exceptions
#
# note: ISO-8859-9E is non standard, ISO-8859-{6,8} are of limited use
# as locales (since all modern handling of bidi is based on utf-8 anyway),
# so they should be removed once UTF-8 is deployed)
EXCEPTION_DIRS = (
    'C', 'POSIX', 'CP1251', 'CP1255', 'CP1256',
    'ISO-8859-1', 'ISO-8859-2', 'ISO-8859-3', 'ISO-8859-4', 'ISO-8859-5',
    'ISO-8859-6', 'ISO-8859-7', 'ISO-8859-8', 'ISO-8859-9', 'ISO-8859-9E',
    'ISO-8859-10', 'ISO-8859-13', 'ISO-8859-14', 'ISO-8859-15',
    'KOI8-R', 'KOI8-U', 'UTF-8', 'default')


def is_valid_lang(lang):
    # TODO: @Foo and charset handling
    lang = re.sub('[@.].*$', '', lang)

    if lang in LANGUAGES:
        return True

    ix = lang.find('_')
    if ix == -1:
        return False

    # TODO: don't accept all lang_COUNTRY combinations
    country = lang[ix + 1:]
    if country not in COUNTRIES:
        return False

    lang = lang[0:ix]
    if lang not in LANGUAGES:
        return False

    return True


class I18NCheck(AbstractCheck):
    def check_binary(self, pkg):
        files = list(pkg.files.keys())
        files.sort()
        locales = []                      # list of locales for this packages
        webapp = False

        i18n_tags = pkg[rpm.RPMTAG_HEADERI18NTABLE] or ()

        for i in i18n_tags:
            try:
                correct = INCORRECT_LOCALES[i]
                self.output.add_info('E', pkg, 'incorrect-i18n-tag-' + correct, i)
            except KeyError:
                pass

        # as some webapps have their files under /var/www/html, and
        # others in /usr/share or /usr/lib, the only reliable way
        # sofar to detect them is to look for an apache configuration file
        for f in files:
            if f.startswith('/etc/apache2/') or \
                    f.startswith('/etc/httpd/conf.d/'):
                webapp = True

        for f in files:
            res = locale_regex.search(f)
            if res:
                locale = res.group(2)
                # checks the same locale only once
                if locale not in locales:
                    locales.append(locale)
                    res2 = correct_subdir_regex.search(locale)
                    if not res2:
                        if locale not in EXCEPTION_DIRS:
                            self.output.add_info('E', pkg, 'incorrect-locale-subdir', f)
                    else:
                        locale_name = res2.group(2)
                        try:
                            correct = INCORRECT_LOCALES[locale_name]
                            self.output.add_info('E', pkg, 'incorrect-locale-' + correct, f)
                        except KeyError:
                            pass
            res = lc_messages_regex.search(f)
            subdir = None
            if res:
                subdir = res.group(1)
                if not is_valid_lang(subdir):
                    self.output.add_info('E', pkg, 'invalid-lc-messages-dir', f)
            else:
                res = man_regex.search(f)
                if res:
                    subdir = res.group(1)
                    if is_valid_lang(subdir):
                        subdir = None
                    else:
                        self.output.add_info('E', pkg, 'invalid-locale-man-dir', f)

            if (f.endswith('.mo') or subdir) and (pkg.files[f].lang == '' and not webapp):
                self.output.add_info('W', pkg, 'file-not-in-%lang', f)

        main_dir, main_lang = ('', '')
        for f in files:
            lang = pkg.files[f].lang
            if main_lang and lang == '' and is_prefix(main_dir + '/', f):
                self.output.add_info('E', pkg, 'subfile-not-in-%lang', f)
            if main_lang != lang:
                main_dir, main_lang = f, lang

        name = pkg.name
        res = package_regex.search(name)
        if res:
            locales = 'locales-' + res.group(1)
            if locales != name and locales not in (x[0] for x in pkg.requires):
                self.output.add_info('E', pkg, 'no-dependency-on', locales)


def is_prefix(p, s):
    return len(p) <= len(s) and p == s[:len(p)]



from mockdata.mock_alternatives import AlternativeConfFolder
import pytest
from rpmlint.checks.AlternativesCheck import AlternativesCheck
from rpmlint.filter import Filter

from Testing import CONFIG, get_tested_package


@pytest.fixture(scope='function', autouse=True)
def alternativescheck():
    CONFIG.info = True
    output = Filter(CONFIG)
    test = AlternativesCheck(CONFIG, output)
    return output, test


#
# udpate-alternatives tests
#


@pytest.mark.parametrize('package', ['binary/alternatives-ok'])
def test_update_alternative_ok(tmp_path, package, alternativescheck):
    output, test = alternativescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'I: package-supports-update-alternatives' in out
    assert 'E' not in out
    assert 'W' not in out


@pytest.mark.parametrize('package', ['binary/alternatives-borked'])
def test_update_alternative_borked(tmp_path, package, alternativescheck):
    output, test = alternativescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: update-alternatives-requirement-missing' in out
    assert 'E: alternative-generic-name-not-symlink' in out
    assert 'E: alternative-link-not-ghost' in out
    assert 'E: update-alternatives-postun-call-missing' in out


@pytest.mark.parametrize('package', ['binary/self'])
def test_non_update_alternative_pkg(tmp_path, package, alternativescheck):
    output, test = alternativescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    # here we just check if there is no requirements checking on
    # non update-alternatived package
    assert 'E' not in out
    assert 'W' not in out


@pytest.mark.parametrize('package', ['binary/python39-evtx'])
def test_update_alternatives_correctness(tmp_path, package, alternativescheck):
    output, test = alternativescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'E: update-alternatives-postun-call-missing' not in out


#
# libalternatives tests
#


@pytest.mark.parametrize('package', ['binary/libalternatives-ok'])
def test_libalternative_ok(tmp_path, package, alternativescheck):
    output, test = alternativescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'I: package-supports-libalternatives' in out
    assert 'E' not in out
    assert 'W' not in out


@pytest.mark.parametrize('package', ['binary/libalternatives-borked'])
def test_libalternative_borked(tmp_path, package, alternativescheck):
    output, test = alternativescheck
    test.check(get_tested_package(package, tmp_path))
    out = output.print_results(output.results)
    assert 'I: package-supports-libalternatives' in out
    assert 'I: libalternatives-conf-not-found' in out
    assert 'E: alts-requirement-missed' in out
    assert 'E: libalternatives-directory-not-exist' in out
    assert 'E: empty-libalternatives-directory' in out
    assert 'W: man-entry-value-not-found' in out
    assert 'W: binary-entry-value-not-found' in out


@pytest.mark.parametrize('package', [AlternativeConfFolder])
def test_alternative_conf_folder(package, alternativescheck):
    output, test = alternativescheck
    test.check(package)
    out = output.print_results(output.results)
    assert 'E: libalternatives-conf-not-found' not in out
    assert 'W: wrong-tag-found' in out
    assert 'E: wrong-entry-format' in out
    assert 'W: binary-entry-value-not-found' in out
